Require Import Hahn Omega Operations Language LangSemantics.
Require Import Setoid Morphisms ClassicalDescription.

Set Implicit Arguments.


(** Implementation environments *)

Definition op_impl_type (op : operation) := 
  vector (op_args op) -> expr (op_result op) nil.

Definition impl_environment := forall op, option (op_impl_type op).

Definition env_of_impl op (body: op_impl_type op) : impl_environment := fun op' =>
  match excluded_middle_informative (op = op') with
  | left EQ => Some (eq_rect _ _ body _ EQ)
  | right _ => None
  end.

Definition env_compose (env1 env2 : impl_environment) : impl_environment :=
  fun op => match env1 op with Some res => Some res | None => env2 op end.

(** Casting return values to larger loop context *)

Fixpoint rt_cast A As (r: return_type A As) Bs : return_type A (As ++ Bs) := 
  match r with 
  | RTbase v => RTbase v
  | RTcons EQ r => RTcons (f_equal (fun x => x ++ Bs) EQ) (rt_cast r Bs)
  end.

(** Casting return values to larger loop context *)

Lemma nth_error_app1 A (l l': list A) n x :
  nth_error l n = Some x ->
  nth_error (l ++ l') n = Some x.
Proof.
  revert l; induction n; ins; desf; ins; desf; eauto.
Qed. 

Fixpoint expr_cast A As (e : expr A As) Bs : expr A (As ++ Bs) := 
  match e with 
  | Evalue v => Evalue v
  | Ecall op args => Ecall op args
  | Elet e1 e2 => Elet (expr_cast e1 _) (fun x => expr_cast (e2 x) _)
  | Eloop e => Eloop (expr_cast e _)
  | Ebreak n v EQ => Ebreak n v (nth_error_app1 _ _ _ EQ)
  | Epar e1 e2 => Epar e1 e2
  end.

(** Substitution of calls by their implementations *)

Fixpoint expand_fcall (env: impl_environment) A As (e: expr A As) : expr A As := 
  match e with 
  | Evalue v => Evalue v
  | Ecall op args => 
      match @env op with
      | None => Ecall op args
      | Some impl => expr_cast (impl args) _
      end
  | Elet e1 e2 => Elet (expand_fcall env e1) (fun x => expand_fcall env (e2 x))
  | Eloop e => Eloop (expand_fcall env e)
  | Ebreak n v EQ => Ebreak n v EQ
  | Epar e1 e2 => Epar (expand_fcall env e1) (expand_fcall env e2)
  end.

(** We now prove a basic lemmas about [expr_cast]. *)

Lemma expr_sem_cast A As (e: expr A As) Bs r gi poi
      (S: expr_sem (expr_cast e Bs) (Some r) gi poi) : 
  exists r0, r = rt_cast r0 Bs /\ expr_sem e (Some r0) gi poi.
Proof.
  revert r gi poi S; induction e; ins;
    unfold idom_seq, idom_loop, idom_loop_helper, idom_par, event_of_call in *; 
    ins; desf; ins.
  - exists (RTbase v); splits; eauto.
  - exists (RTbase (ev_res a)); splits; eauto.
    by left; exists a; eexists; splits; ins; exists eq_refl; ins. 
  - eapply IHe in S; eapply H in S0; desf.
    destruct r1; ins; desf; eauto 15.
  - eapply IHe in S; desf.
    destruct r0; ins; desf; eexists (RTcons eq_refl r0); split; eauto 8.
    by ins; f_equal; apply UIP_refl.
  - eapply IHe in LAST; eauto; desf.
    destruct r0; ins; desf.
    assert (B0 = A /\ Bs1 = As); desf.
      by clear - EQ0; desf.
    eexists r0; split. 
      by rewrite <- !eq_rect_eq.
    do 5 eexists; splits; ins; eauto; ins; rewrite <- ?eq_rect_eq; ins.
    eapply NLAST, IHe in LT; desf; destruct r1; ins; desf.
  - eexists (RTconsn n v EQ); split; eauto.
    clear; generalize (nth_error_app1 As Bs n EQ) as PF; revert B As EQ. 
    induction n; destruct As; ins; f_equal; eauto.
    assert (T = A); [by desf| subst; ins].
    by apply Streicher_K, UIP_refl.  
  - destruct r1, r2; ins; desf.
    exists (RTbase (v, v0)); split; ins; eauto 15.
Qed.

Lemma expand_fcall_comp (env1 env2 : impl_environment) 
      (D: forall op, env1 op <> None -> env2 op <> None -> False)
      (ID: forall op e (EQ: env1 op = Some e) args As, 
          expand_fcall env2 (expr_cast (e args) As) = expr_cast (e args) As)
      A As (e: expr A As) :
  expand_fcall (env_compose env1 env2) e = expand_fcall env2 (expand_fcall env1 e).
Proof.
  induction e; ins; desf. 
  all: try unfold env_compose in Heq; desf; ins; desf; 
       try congruence; eauto using eq_sym.
  f_equal; try congruence; apply functional_extensionality; ins. 
Qed.

Definition plainexec_related (env : impl_environment) (gi gs : event -> Prop) (poi pos : relation event) := 
  exists (fs: event -> event),
    << gS : gs ≡₁ set_collect fs gi >> /\
    << poS: pos ≡ collect_rel fs poi \ <| fun _ => True |> >> /\
    << poWF: forall x y (PO: poi x y) (NEQ: fs x <> fs y) z (Gz: gi z)
                        (EQ1: fs z = fs x) w (Gw: gi w )(EQ2: fs w = fs y), poi z w >> /\
    << opEQ: forall x (Gx: gi x) (NENV: env (ev_op (fs x)) = None), fs x = x >> /\
    << INJ: forall x (Gx: gi x) y (Gy: gi y) (EQ: fs x = fs y) 
                     (ENV: env (ev_op (fs x)) = None), x = y >> /\
    << ESEM:  forall a (Ga: gs a) body (ENV: env (ev_op a) = Some body), 
             expr_sem (body (ev_args a)) (Some (RTbase (ev_res a))) 
                      (gi ∩₁ (fun x => fs x = a))
                      (restr_rel (fun x => fs x = a) poi) >>.

Add Parametric Morphism A As : (@expr_sem A As) with signature 
  eq ==> eq ==> set_equiv ==> same_relation ==> iff as expr_sem_more.
Proof.
  induction y; ins; unfold idom_seq, idom_loop, idom_par in *;
    repeat first [apply and_iff_compat_l | apply and_iff_compat_r |
                  apply Coq.Classes.Morphisms_Prop.or_iff_morphism | 
                  apply Coq.Classes.Morphisms_Prop.ex_iff_morphism; intro ];
    unnw; ins; eauto.
  all: rewrite ?H, ?H0, ?H1; ins.
Qed.

Add Parametric Morphism : plainexec_related with signature 
  eq ==> set_equiv ==> set_equiv ==> same_relation ==> same_relation ==> iff as plainexec_related_more.
Proof.
  unfold plainexec_related; ins; unnw;
    repeat first [match goal with |- _ /\ _ <-> _ => 
                    apply Coq.Classes.Morphisms_Prop.and_iff_morphism end |
                  apply Coq.Classes.Morphisms_Prop.or_iff_morphism | 
                  apply Coq.Classes.Morphisms_Prop.ex_iff_morphism; intro ];
    rewrite ?H, ?H0, ?H1, ?H2; ins; 
    split; ins; rewrite ?H, ?H0, ?H1, ?H2; ins; eauto; try solve [firstorder].
  rewrite <- ?H, <- ?H0, <- ?H1, <- ?H2; eapply H3; eauto; firstorder. 
Qed.

Lemma plainexec_related_par env g1 gs1 po1 pos1
  (R1 : plainexec_related env g1 gs1 po1 pos1) g2 gs2 po2 pos2
  (R2 : plainexec_related env g2 gs2 po2 pos2)
  (D1 : po1 ⊆ g1 × g1)  (D2 : po2 ⊆ g2 × g2)
  (Ds1 : pos1 ⊆ gs1 × gs1)  (Ds2 : pos2 ⊆ gs2 × gs2)
  (DISJ : set_disjoint g1 g2) (DISJs : set_disjoint gs1 gs2) :
  plainexec_related env (g1 ∪₁ g2) (gs1 ∪₁ gs2) (po1 ∪ po2) (pos1 ∪ pos2).
Proof.
  destruct R1 as (fs1 & R); desc. 
  destruct R2 as (fs2 & R); desc. 
  rewrite gS, gS0, poS, poS0 in *; clear poS poS0. 
  exists (fun x => if excluded_middle_informative (g1 x) then fs1 x else fs2 x); splits; ins; relsf.
  - unfolder; split; ins; desf; try solve [left; exists y; desf|right; exists y; desf; firstorder].
  - relsf; rewrite <- minus_union_l; apply minus_rel_more; ins.
    apply union_more; unfolder; split; ins; desf; splits; ins; clarify_not; exists x', y'; splits; ins;
    try apply D1 in H; try apply D2 in H; red in H; desf; firstorder.
  - unfold set_disjoint in DISJ, DISJs; unfold set_collect in DISJs.
    clear ESEM INJ opEQ gS ESEM0 INJ0 opEQ0 gS0. 
    destruct PO; [left; specialize (D1 _ _ H); red in D1|
                  right;specialize (D2 _ _ H); red in D2]; desf; eauto 2;
      try solve [exfalso; eauto 2].
    all: red in Gz; red in Gw; desf; eauto 2; exfalso; eauto 2.
    all: match goal with H: _ = ?xx |- _ => solve [eapply (DISJs xx); eauto] end.  
  - red in Gx; desf; eauto.
  - red in Gx, Gy; desf; eauto. 
    all: exfalso; eapply DISJs; eexists; eauto.
  - assert (CONTRA: forall x y, g1 x -> g2 y -> fs1 x = fs2 y -> False).
      by clear - DISJs; firstorder.
    clear DISJs; specialize (ESEM a); specialize (ESEM0 a). 
    unfolder in Ga; desf; [clear ESEM0|clear ESEM gS; rename ESEM0 into ESEM; rename gS0 into gS].
    all: specialize_full ESEM; eauto; [by apply gS; vauto|].
    all: eapply expr_sem_more, ESEM; eauto; 
         [unfolder; splits; ins; desf; eauto; exfalso; eauto; firstorder|].
    all: clear - D1 D2 DISJ Ga CONTRA.
    all: rewrite restr_union.
    all: unfolder; splits; ins; desf; eauto.
    all: exfalso; try first [apply D1 in H | apply D2 in H]; 
         red in H; desc; eauto using eq_sym; firstorder.
Qed.


Lemma plainexec_related_seq env g1 gs1 po1 pos1
  (R1 : plainexec_related env g1 gs1 po1 pos1) g2 gs2 po2 pos2
  (R2 : plainexec_related env g2 gs2 po2 pos2)
  (D1 : po1 ⊆ g1 × g1)  (D2 : po2 ⊆ g2 × g2)
  (Ds1 : pos1 ⊆ gs1 × gs1)  (Ds2 : pos2 ⊆ gs2 × gs2)
  (DISJ : set_disjoint g1 g2) (DISJs : set_disjoint gs1 gs2) :
  plainexec_related env (g1 ∪₁ g2) (gs1 ∪₁ gs2) (po1 ∪ po2 ∪ g1 × g2) (pos1 ∪ pos2 ∪ gs1 × gs2).
Proof.
  destruct R1 as (fs1 & R); desc. 
  destruct R2 as (fs2 & R); desc. 
  rewrite gS, gS0, poS, poS0 in *; clear poS poS0. 
  exists (fun x => if excluded_middle_informative (g1 x) then fs1 x else fs2 x); splits; ins; relsf.
  - unfolder; split; ins; desf; try solve [left; exists y; desf|right; exists y; desf; firstorder].
  - { rewrite !minus_union_l; repeat apply union_more; try apply minus_rel_more; ins.
      1-2: unfolder; split; ins; desf; splits; ins; clarify_not; exists x', y'; splits; ins;
      try apply D1 in H; try apply D2 in H; red in H; desf; firstorder.
      unfolder; split; ins; desf; splits; try red; ins; desf; eauto.
      1-2: eexists; split; eauto; desf; eauto.
      2: eapply DISJs; eexists; eauto.
      all: exfalso; firstorder. } 
  - unfold set_disjoint in DISJ, DISJs; unfold set_collect in DISJs.
    clear ESEM INJ opEQ gS ESEM0 INJ0 opEQ0 gS0. 
    destruct PO as [[PO|PO]|PO]; [left;left; specialize (D1 _ _ PO); red in D1|
                  left;right;specialize (D2 _ _ PO); red in D2|
                  right; red; red in PO]; desf; eauto 2;
      try solve [exfalso; eauto 2].
    all: red in Gz; red in Gw; desf; eauto 2; try solve [exfalso; eauto 2].
    all: try match goal with H: _ = ?xx |- _ => solve [destruct (DISJs xx); eauto] end.  
  - red in Gx; desf; eauto.
  - red in Gx, Gy; desf; eauto; try by exfalso; eapply (DISJs); eexists; eauto.
  - assert (CONTRA: forall x y, g1 x -> g2 y -> fs1 x = fs2 y -> False).
      by clear - DISJs; firstorder.
    specialize (ESEM a); specialize (ESEM0 a). 
    unfolder in Ga; desf; [clear ESEM0|clear ESEM gS; rename ESEM0 into ESEM; rename gS0 into gS].
    all: specialize_full ESEM; eauto; [by apply gS; vauto|].
    all: eapply expr_sem_more, ESEM; eauto; 
         [unfolder; splits; ins; desf; eauto; exfalso; eauto; firstorder|].
    all: clear - D1 D2 DISJ Ga CONTRA.
    all: unfolder; splits; ins; desf; eauto.
    all: exfalso; try first [apply D1 in H | apply D2 in H]; try red in H; desc; 
         eauto using eq_sym; firstorder.
Qed.


Theorem expand_related env 
      (NEMP : forall op body args r gi poi, 
          env op = Some body -> expr_sem (body args) (Some r) gi poi -> ~ gi ≡₁ ∅)  
      A As (e: expr A As) : 
  forall r gi poi (IMPL: expr_sem (expand_fcall env e) (Some r) gi poi),
    exists gs pos, 
     <<ESEM: expr_sem e (Some r) gs pos >> /\ 
     <<INCL: set_collect ev_id gs ⊆₁ set_collect ev_id gi >> /\
     <<REL: plainexec_related env gi gs poi pos >>.
Proof.
  induction e; ins; unfold idom_seq, idom_par in *; ins; desf; ins; desf.
  { eexists ∅, ∅₂; splits; rels; exists (fun x => x); splits; ins; firstorder. }
  {
    eapply expr_sem_cast with (e := e args) in IMPL; desf.
    destruct r0; ins.
    specialize_full NEMP; eauto.
    rewrite set_nonemptyE in NEMP; destruct NEMP as (iev & Giev).
    set (a := {| ev_id := ev_id iev; ev_op := op; ev_args := args ; ev_res := v |}).
    exists (eq a), ∅₂; splits; rels. 
      by left; eexists a; eexists; splits; ins; exists eq_refl; rewrite <- eq_rect_eq; splits; ins.
      by unfolder; intuition; desf; eauto. 
    exists (fun _ => a); splits; ins; desf.
      by unfolder; intuition; desf; eauto.
      by clear; unfolder; intuition; desf; eauto.
    arewrite ((fun _ : event => a = a) ≡₁ (fun _ => True)); rewrite restr_relE; rels.
    subst a; ins; desf.
  }
  { exists (eq a), ∅₂; splits; rels; eauto 7.
      by rewrite gEQ; unfolder; intuition; desf; eauto. 
    red in aCALL; desf.
    exists (fun x => a); splits; ins; desf; rewrite ?gEQ; unfolder in *;
      intuition; desf; eauto; ins; desf.
      by apply H in Gx; apply H in Gy; desf. 
  }
  (* Let *)
  { specialize_full IHe; eauto; destruct IHe as (gs1 & pos1 & ?); desc.
    specialize_full H; eauto; destruct H as (gs2 & pos2 & ?); desc.
    assert (DISJOINT: set_disjoint (set_collect ev_id gs1) (set_collect ev_id gs2)).
      by red; ins; eauto. 
    do 2 eexists; splits; [by eauto 15| |]. 
      by rewrite gEQ, !set_collect_union, INCL, INCL0. 
    rewrite gEQ, poEQ; eapply plainexec_related_seq; eauto using esem_dom, set_collect_disjointD.
  }
  (* Let / break *)
  { specialize_full IHe; eauto; desc; do 2 eexists; split; eauto 10. }
  (* Loop *) 
  {
    red in IMPL; desf.
    assert (exists (gsn : nat -> event -> Prop) posn, 
          ⟪SEM: forall i (LT: i < n), expr_sem e (Some (RTbase tt)) (gsn i) (posn i)⟫ /\
          ⟪INCL: forall i (LT: i < n), set_collect ev_id (gsn i) ⊆₁ set_collect ev_id (gn i)⟫ /\
          ⟪REL: plainexec_related env (⋃₁ i < n, gn i) (⋃₁ i < n, gsn i) 
                                      (⋃ i < n, pon i ∪ (⋃ j < i, gn j × gn i))
                                      (⋃ i < n, posn i ∪ (⋃ j < i, gsn j × gsn i))⟫).
    {
      clear - gCOMPAT IHe NLAST; induction n.
      exists (fun _ => ∅), (fun _ => ∅₂); splits; ins; try omega.
        unfolder; intuition; desf.
        exists (fun x => x); splits; unfolder; ins; intuition; desf; omega.
      specialize_full IHn; ins; eauto with arith; desf.
      assert (NL' := NLAST); specialize (NLAST n); specialize_full NLAST; try omega.
      eapply IHe in NLAST; eauto; desf.
      exists (fun i => if Nat.eq_dec i n then gs else gsn i).
      exists (fun i => if Nat.eq_dec i n then pos else posn i).
      splits; ins; desf; eauto; try omega.
      - eapply SEM; omega.
      - apply INCL; omega. 
      - assert (X: (fun i => i < S n) ≡₁ (fun i => i < n) ∪₁ (eq n)). 
          by clear; unfolder; intuition; omega.
        rewrite X; relsf; rewrite ?bunion_union_l, ?bunion_eq; desf; desf.
        eapply plainexec_related_more, plainexec_related_seq with 
            (g1 := ⋃₁ i < n, gn i) (g2 := gn n); try eassumption; 
          unionL; eauto using esem_dom.
        * apply set_equiv_union; ins; apply set_equiv_bunion; ins; desf; omega.
        * rewrite !unionA; do 2 (apply union_more; ins); clear; unfolder; firstorder.
        * rewrite !unionA; repeat apply union_more; ins.
          2: clear; unfolder; intuition; desf; try omega; eauto; eexists; split; eauto; desf; omega.
          clear; unfolder; intuition; desf; try omega; eauto; eexists; split; eauto; desf; eauto; try omega.
          right; eexists; split; eauto; desf; omega.
        * specialize (NL' x); specialize_full NL'; eauto with arith.
          apply esem_dom in NL'; rewrite NL'; clear - H.
          by unfolder; ins; desf; eauto 6. 
        * by unfolder; ins; desf; split; eauto with arith.
        * assert (Y:=H); apply SEM, esem_dom in Y; rewrite Y.
          by unfolder; ins; desf; eauto 6. 
        * by unfolder; ins; desf; split; eauto with arith.
        * clear -gCOMPAT COND; unfolder in *; ins; desf; eauto. 
          by eapply gCOMPAT; eauto 10.
        * clear -gCOMPAT INCL INCL0 COND; unfolder in *; ins; desf; eauto. 
          by eapply gCOMPAT; eauto 8.
    }
    destruct rt as [rr|]; try done.
    desf; specialize_full IHe; eauto; desf.
    exists ((⋃₁ i < n, gsn i) ∪₁ gs), 
           ((⋃ i < n, posn i ∪ (⋃ j < i, gsn j × gsn i)) ∪ pos ∪ (⋃₁ i < n, gsn i) × gs);
      splits.
     exists n, (Some rr). 
     exists (fun i => if Nat.eq_dec i n then gs else gsn i).
     exists (fun i => if Nat.eq_dec i n then pos else posn i); splits; ins; desf; eauto; try omega.
     - clear - gCOMPAT LT LEQ INCL INCL0; unfold set_subset, set_disjoint in *; ins; desf. 
       by eapply gCOMPAT; eauto. 
     - clear - gCOMPAT LT LEQ INCL INCL0 n0 n1; unfold set_subset, set_disjoint in *; ins; desf. 
       eapply INCL in IN; try omega.
       eapply INCL in IN'; try omega; eauto. 
     - rewrite set_le; relsf; desf; desf; apply set_equiv_union; ins. 
       apply set_equiv_bunion; ins; desf; omega.
     - {
         rewrite set_le, bunion_union_l, bunion_eq, unionA; desf; desf; repeat apply union_more; ins. 
         all: clear; unfolder; intuition; desf; eauto 8; try omega.
         exists a; desf; desf; split; eauto; omega. 
         exists a; desf; desf; split; eauto; try omega; right; eexists; split; eauto; desf; omega.
         exists y0; desf; omega.
       }
     - rewrite gEQ, set_le, set_bunion_union_l, set_bunion_eq, !set_collect_union.
       eapply set_subset_union; ins.
       clear -INCL; unfolder in *; ins; desf.
       forward eapply INCL with (x := ev_id y) as X; eauto; desf; eauto.
     - rewrite gEQ, poEQ, set_le, bunion_union_l, bunion_eq, <- unionA; relsf.
       eapply plainexec_related_more, plainexec_related_seq with 
         (g1 := ⋃₁ i < n, gn i) (g2 := gn n); try eassumption; eauto using esem_dom;
        unionL; ins.
        * repeat apply union_more; ins; clear; unfolder; intuition; desf; eauto.
        * assert (Y:=H); apply NLAST, esem_dom in Y; rewrite Y.
          by unfolder; ins; desf; eauto 6.
        * by unfolder; ins; desf; split; eauto with arith.
        * assert (Y:=H); apply SEM, esem_dom in Y; rewrite Y.
          by unfolder; ins; desf; eauto 6.
        * by unfolder; ins; desf; split; eauto with arith.
        * clear -gCOMPAT COND; unfolder in *; ins; desf; eauto. 
          by eapply gCOMPAT; eauto 8.
        * clear -gCOMPAT COND INCL INCL0; unfolder in *; ins; desf; eauto. 
          by eapply gCOMPAT; eauto 8.
   }
  (* Break *)
  { eexists ∅, ∅₂; splits; rels; exists (fun x => x); splits; ins; firstorder. }
  (* Parallel *)
  { unfold idom_par_helper; destruct r1, r2; ins; desf.
    specialize_full IHe1; eauto; destruct IHe1 as (gs1 & pos1 & ?); desc.
    specialize_full IHe2; eauto; destruct IHe2 as (gs2 & pos2 & ?); desc.
    assert (DISJOINT: set_disjoint (set_collect ev_id gs1) (set_collect ev_id gs2)).
      by red; eauto. 
    do 2 eexists; split; [by eauto 15|]; splits.
      by rewrite gEQ, !set_collect_union, INCL, INCL0. 
    rewrite gEQ, poEQ; apply plainexec_related_par; eauto using esem_dom, set_collect_disjointD. 
  }
Qed.

