Require Import Hahn Operations Language. 

Set Implicit Arguments.

Inductive return_type (A: Type) (As: list Type) : Type := 
  | RTbase (v : A) 
  | RTcons B Bs (EQ: As = B::Bs) (rv : return_type B Bs).

Arguments RTbase [A As] v.
Arguments RTcons [A As B Bs] EQ rv.

Program Fixpoint RTconsn (n : nat) {A As} (x : A) (EQ: nth_error As n = Some A) {B} : 
  return_type B As := 
  match n with 
  | 0 => RTcons (Bs := tl As) _ (RTbase x)
  | S n => RTcons (B := hd A As) (Bs := tl As) _ (RTconsn n x _)
  end.
Solve Obligations with (ins; subst; simpls; desf; desf). 

Definition interp_dom (A: Type) (As: list Type) :=
  option (return_type A As) -> (event -> Prop) -> relation event -> Prop.

Definition interp_doml (As: list Type) :=
  option (vector As) -> (event -> Prop) -> relation event -> Prop.

Definition domrange A (r : relation A) := dom_rel r ∪₁ codom_rel r.

(** Method calls *)

Definition event_of_call op (args : vector (op_args op)) (res : op_result op) (a : event) :=
  exists (OP: ev_op a = op), 
    ⟪ARGS: ev_args a = eq_rect _ _ args _ (f_equal _ (eq_sym OP))⟫
    /\ ⟪RES: res = eq_rect _ _ (ev_res a) _ OP⟫.

(** Sequential composition *)

Definition idom_seq A As (x : interp_dom A As) B (y : A -> interp_dom B As) : 
  interp_dom B As := 
  fun r g po => 
       (exists v g1 po1 g2 po2, x (Some (RTbase v)) g1 po1 /\ y v r g2 po2 
         /\ ⟪gDISJ: set_disjoint (set_collect ev_id g1) (set_collect ev_id g2)⟫ 
         /\ ⟪gEQ: g ≡₁ g1 ∪₁ g2⟫ /\ ⟪poEQ: po ≡ po1 ∪ po2 ∪ g1 × g2⟫)
    \/ (exists C Cs (EQ : As = C :: Cs) (rt : return_type C Cs), 
           x (Some (RTcons EQ rt)) g po /\ r = Some (RTcons EQ rt))
    \/ r = None /\ x None g po.

Lemma cons_inj A (x : A) y z w : x :: y = z :: w -> x = z /\ y = w.
Proof. ins; desf. Qed.

(** Loops *)

Definition idom_loop_helper A As (x: option (return_type unit (A::As)))
  : option (return_type A As) := 
  match x with 
  | Some (@RTcons _ _ B Bs EQ rt) => 
      let mEQ := eq_sym (proj1 (cons_inj EQ)) in
      let nEQ := eq_sym (proj2 (cons_inj EQ)) in
      Some (eq_rect _ (fun As => return_type A As)
                    (eq_rect _ (fun A => return_type A Bs) rt _ mEQ) 
                    _ nEQ)
  | _ => None
  end.

Definition idom_loop A As (x : interp_dom unit (A::As)) : interp_dom A As := 
  fun r g po => 
    exists (n: nat) rt gn pon,
      ⟪NLAST: forall i (LT: i < n), x (Some (RTbase tt)) (gn i) (pon i)⟫ /\
      ⟪LAST: x rt (gn n) (pon n)⟫ /\
      ⟪rEQ : r = idom_loop_helper rt⟫ /\
      ⟪gCOMPAT : forall i j (LT: i < j) (LEQ: j <= n), 
                   set_disjoint (set_collect ev_id (gn i)) (set_collect ev_id (gn j))⟫ /\
      ⟪gEQ : g ≡₁ ⋃₁ i <= n, gn i⟫ /\
      ⟪poEQ : po ≡ ⋃i <= n, pon i ∪ ⋃j < i, gn j × gn i⟫.

(** Parallel composition *)

Definition idom_par_helper A (x: option (return_type A nil)) B (y: option (return_type B nil)) As :
  option (return_type (A * B) As) := 
  match x, y with Some (RTbase a), Some (RTbase b) => Some (RTbase (a, b)) | _, _ => None end.

Definition idom_par A1 (x : interp_dom A1 nil) A2 (y : interp_dom A2 nil) {As}
  : interp_dom (A1 * A2) As := fun r g po => 
      exists r1 g1 po1 r2 g2 po2, 
        x r1 g1 po1 /\ y r2 g2 po2
        /\ ⟪gDISJ: set_disjoint (set_collect ev_id g1) (set_collect ev_id g2)⟫
        /\ ⟪gEQ: g ≡₁ g1 ∪₁ g2⟫
        /\ ⟪poEQ: po ≡ po1 ∪ po2⟫
        /\ ⟪rEQ: r = idom_par_helper r1 r2 As⟫.

(** Expression semantics *)

Fixpoint expr_sem A As (e: expr A As) : interp_dom A As := 
  match e with 
  | Evalue v => fun r g po => 
     ⟪rEQ: r = Some (RTbase v) \/ r = None⟫ /\ ⟪gEQ: g ≡₁ ∅⟫ /\ ⟪poEQ: po ≡ ∅₂⟫
  | Ecall op args => fun r g po =>
     (exists a res,
          ⟪rEQ: r = Some (RTbase res)⟫
       /\ ⟪gEQ: g ≡₁ eq a⟫
       /\ ⟪poEQ: po ≡ ∅₂⟫
       /\ ⟪aCALL: event_of_call op args res a⟫)
     \/ ⟪rEQ: r = None⟫ /\ ⟪gEQ: g ≡₁ ∅⟫ /\ ⟪poEQ: po ≡ ∅₂⟫
  | Elet e1 e2 => idom_seq (expr_sem e1) (fun x => expr_sem (e2 x))
  | Eloop e => idom_loop (expr_sem e) 
  | Ebreak n v EQ => fun r g po => 
     ⟪rEQ: r = Some (RTconsn n v EQ) \/ r = None⟫ /\ ⟪gEQ: g ≡₁ ∅⟫ /\ ⟪poEQ: po ≡ ∅₂⟫
  | Epar e1 e2 => idom_par (expr_sem e1) (expr_sem e2)
  end.

Lemma doma_subset A r (d d' : A -> Prop) : doma r d -> d ⊆₁ d' -> doma r d'.
Proof. unfolder; firstorder. Qed.

Lemma domb_subset A r (d d' : A -> Prop) : domb r d -> d ⊆₁ d' -> domb r d'.
Proof. unfolder; firstorder. Qed.

Lemma doma_bunion A P B (rr: A -> relation B) d : 
  doma (⋃x ∈ P, rr x) d <-> forall i (COND: P i), doma (rr i) d.
Proof. unfolder; firstorder. Qed.

Lemma domb_bunion A P B (rr: A -> relation B) d : 
  domb (⋃x ∈ P, rr x) d <-> forall i (COND: P i), domb (rr i) d.
Proof. unfolder; firstorder. Qed.

(** We now establish some basic properties of the semantics. 
First, for all plain executions generating by a program, [gpo ⊆ g × g]. *)

Lemma esem_doma A As (e: expr A As) :
  forall r g gpo (S: expr_sem e r g gpo), doma gpo g.
Proof.
  induction e; ins; unfold idom_seq, idom_loop, idom_par in *; ins; desf.
  all: 
    repeat match goal with [H : _ ≡ _ |- _] => rewrite H end;
    repeat match goal with H : _ ≡₁ _ |- _ => rewrite H end; ins;
    relsf; repeat (apply union_doma || apply doma_bunion; ins); 
      eauto using doma_subset with hahn.
  apply Lt.le_lt_or_eq in COND; desf; eapply doma_subset; eauto; unfolder; eauto with arith.
  unfolder; ins; desf; eauto with arith.
Qed.

Lemma esem_domb A As (e: expr A As) :
  forall r g gpo (S: expr_sem e r g gpo), domb gpo g.
Proof.
  induction e; ins; unfold idom_seq, idom_loop, idom_par in *; ins; desf.
  all: 
    repeat match goal with [H : _ ≡ _ |- _] => rewrite H end;
    repeat match goal with H : _ ≡₁ _ |- _ => rewrite H end; ins;
    relsf; repeat (apply union_domb || apply domb_bunion; ins); 
      eauto using domb_subset with hahn.
  apply Lt.le_lt_or_eq in COND; desf; eapply domb_subset; eauto; unfolder; eauto with arith.
  unfolder; ins; desf; eauto with arith.
Qed.

Lemma esem_dom A As (e: expr A As) r g gpo (S: expr_sem e r g gpo) : gpo ⊆ g × g.
Proof.
  eapply domab_helper2; eauto using esem_doma, esem_domb.
Qed.

(** In every plain execution of an expression, the program order is transitive. *)

Lemma set_collect_disjointD A B (f: A -> B) s s' : 
  set_disjoint (set_collect f s) (set_collect f s') -> set_disjoint s s'.
Proof.
  unfolder; intuition; eapply H; eauto. 
Qed.

Lemma esem_transitive A As (e: expr A As) :
  forall r g gpo (S: expr_sem e r g gpo), transitive gpo.
Proof.
  induction e; ins; unfold idom_seq, idom_loop, idom_par in *; ins; desf.
  all: 
    repeat match goal with [H : _ ≡ _ |- _] => rewrite H; clear H end;
    repeat match goal with [H : _ ≡₁ _ |- _] => rewrite H; clear H end; ins;
    relsf.
  {
    specialize_full H; eauto.
    specialize_full IHe; eauto.
    apply set_collect_disjointD in gDISJ.
    apply transitiveI; relsf; repeat apply inclusion_union_l; autos.
    all: apply inclusion_union_r; right.
    all: rewrite ?(esem_dom _ _ _ _ S), ?(esem_dom _ _ _ _ S0); firstorder. 
  }
  { 
    assert (IH': forall i (LEQ: i <= n), transitive (pon i)).
      by ins; apply Lt.le_lt_or_eq in LEQ; desf; eauto. 
    assert (X: forall i (LEQ: i <= n),  pon i <<= gn i × gn i). 
      by ins; apply Lt.le_lt_or_eq in LEQ; desf; eauto using esem_dom. 
    clear IHe LAST NLAST gEQ.
    generalize (fun i j LE LT => set_collect_disjointD (gCOMPAT i j LE LT)); 
      clear gCOMPAT; intro gCOMPAT.
    assert (C : forall i j, i <> j -> i <= n -> j <= n -> set_disjoint (gn i) (gn j)).
      by ins; rewrite NPeano.Nat.lt_gt_cases in *; desf; eauto; rewrite set_disjointC; eauto.
    clear gCOMPAT rt.
    apply transitiveI; repeat (relsf; unionL). 
    - destruct (NPeano.Nat.eq_dec x0 x); desf. 
        specialize_full IH'; eauto; relsf; firstorder.
      specialize_full C; eauto.
      rewrite X, X; ins; apply inclusion_trans with (y := ∅₂); ins; firstorder. 
    - eapply inclusion_bunion_r with x0; eauto.
      eapply inclusion_union_r; right.
      eapply inclusion_bunion_r with x1; eauto.
      rewrite X; ins.
      destruct (NPeano.Nat.eq_dec x0 x); desf. firstorder.
      specialize_full C; eauto; firstorder.
    - eapply inclusion_bunion_r with x; eauto.
      eapply inclusion_union_r; right.
      eapply inclusion_bunion_r; eauto.
      rewrite X; ins.
      destruct (NPeano.Nat.eq_dec x0 x1); desf. firstorder.
      specialize_full C; eauto; firstorder.
    - eapply inclusion_bunion_r with x; eauto.
      eapply inclusion_union_r; right.
      destruct (NPeano.Nat.eq_dec x0 x1); desf. firstorder.
      specialize_full C; eauto with arith; firstorder.
  }
  {
    specialize_full IHe1; eauto.
    specialize_full IHe2; eauto.
    apply set_collect_disjointD in gDISJ.
    apply transitiveI; relsf; repeat apply inclusion_union_l; relsf.
    all: apply inclusion_trans with (y := ∅₂); ins.
    all: rewrite ?(esem_dom _ _ _ _ S), ?(esem_dom _ _ _ _ S0); firstorder. 
  }
Qed.

Lemma esem_finite A As (e: expr A As) : 
  forall r g po (S: expr_sem e (Some r) g po), set_finite g.
Proof.
  induction e; ins; unfold idom_seq, idom_loop, idom_par in *; ins; desf.
  all: 
    repeat match goal with [H : _ ≡₁ _ |- _] => rewrite H; clear H end; 
    ins; rels; eauto with hahn.
  2: by destruct r1, r2; ins; desf; eauto.
  destruct rt; ins; desf.
  apply set_finite_bunion; ins; autos.
  rewrite NPeano.Nat.lt_eq_cases in COND; desf; eauto.
Qed.

Lemma esem_fsupp A As (e: expr A As) : 
  forall r g po (S: expr_sem e r g po), fsupp po.
Proof.
  induction e; ins; unfold idom_seq, idom_loop, idom_par in *; ins; desf.
  all: 
    repeat match goal with [H : _ ≡₁ _ |- _] => rewrite H; clear H end; 
    repeat match goal with [H : _ ≡ _ |- _] => rewrite H; clear H end; 
    repeat first [eapply fsupp_bunion; intros | apply fsupp_unionI ]; 
    eauto using esem_finite with hahn.
  rewrite NPeano.Nat.lt_eq_cases in IN; desf; eauto with hahn.
  eauto using esem_finite with arith hahn.
Qed.
