Require Import String List ZArith Omega Operations.

Set Implicit Arguments.

(** ** Program expressions

Operations contain a name, a list of types for the arguments, 
and a type for the result. *)

Inductive expr : Type -> list Type -> Type := 
  | Evalue A (v : A) As : expr A As
  | Ecall (op : operation) (args: vector (op_args op)) 
          As : expr (op_result op) As
  | Elet B As (e1 : expr B As) A (e2 : B -> expr A As) : expr A As
  | Eloop A As (e : expr unit (A::As)) : expr A As 
  | Ebreak (n : nat) A (v: A) As 
           (EQ: nth_error As n = Some A) B : expr B As
  | Epar A1 (e1 : expr A1 nil) A2 (e2: expr A2 nil) As : expr (A1 * A2) As.

Arguments Evalue [A] v [As].
Arguments Ecall op args [As].
Arguments Ebreak n [A] v [As] EQ [B].
Arguments Epar [A1] e1 [A2] e2 [As].

Definition Eseq A As (e1 : expr unit As) (e2 : expr A As) :=
  Elet e1 (fun _ => e2).

Definition BREAK {A As} : expr A ((unit:Type)::As) := 
  Ebreak (As:=(unit:Type)::As) 0 tt eq_refl.

Definition BREAK_RET {A B As} (v : B) : expr A (B::As) := 
  Ebreak (As:=B::As) 0 v eq_refl.

Definition BREAK_RET2 {A C B As} (v : B) : expr A (C::B::As) := 
  Ebreak (As:=C::B::As) 1 v eq_refl.

Program Definition Erepeat (As: list Type) (e : expr bool (unit::As)) 
  : expr unit As :=
  Eloop (Elet e (fun x => if x then BREAK else Evalue tt)).


(** Notations for easier encoding of programs. *)
Notation " 'LET' x <- E 'in' F " := 
  (Elet E  (fun x => F)) 
  (at level 200, x ident). 
Notation " 'LOOP' E 'END' " := (Eloop E).
Notation " 'REPEAT' E 'END' " := (Erepeat E).
(*Notation " 'PAR_BEGIN' E 'PAR' F 'END' " := (Epar E F).
Notation " {{ E }}_ A " := (Eload A E) (at level 198, A ident).
Notation " {{ E }}_ A <- V " := (Estore A E V) (at level 198, A ident). *)
Notation " E ;;; F " := (Eseq E F) (at level 199). 

Notation "'CALL' E ( ) " := 
  (Ecall E tt)
    (E at level 0). 

Notation "'CALL' E ( F , .. , G ) " := 
  (Ecall E (pair F .. (pair G tt) .. )) 
    (E at level 0). 


(**************************************************************************)
(** ** C11 atomic operations *)
(**************************************************************************)

Local Open Scope Z. 

(**************************************************************************)
(** ** For loops *)
(**************************************************************************)

Definition Efor As (min max: Z) (body: Z -> expr unit ((unit:Type)::As)) : expr unit As := 
  LET ivar <- CALL reg_alloc ( ) in
  CALL reg_write (ivar, min) ;;;
  LOOP
    LET i <- CALL reg_read (ivar) in
    if i <=? max then body i else BREAK
  END.

Notation "'FOR' x <- E 'TO' F 'DO' G 'END' " := (Efor E F (fun x => G)) (x ident).

(**************************************************************************)
(** ** Fetch_add and atomic_swap *)
(**************************************************************************)
(*
Definition fadd_body : my_op_impl_type C11_fadd := fun addr diff mode =>
  LOOP
    LET old <- CALL C11_load (addr , RMrlx) in
    LET b <- CALL C11_cas (addr , old , old + diff , mode , RMrlx) in 
    if b then BREAK_RET old else Evalue tt
  END.

Definition xchg_body : my_op_impl_type C11_xchg := fun addr new mode =>
  LOOP
    LET old <- CALL C11_load (addr , RMrlx) in
    LET b <- CALL C11_cas (addr , old , new , mode , RMrlx) in 
    if b then BREAK_RET old else Evalue tt
  END.
*)
(**************************************************************************)
(** ** Mutual exclusion locks *)
(**************************************************************************)

(** A simple spinlock implementation: 
  - To acquire the lock, we do an acquire CAS changing its value from 0 to 1.
  - To release the lock, we perform a release store with value 0. 
*)
(*
Definition initlock_body : my_op_impl_type op_initlock := fun addr =>
  CALL C11_store (addr , 0 , WMrlx).

Definition newlock_body : my_op_impl_type op_newlock :=
  LET block <- CALL alloc_block ( ) in 
  CALL C11_store ((block,0%nat) , 0 , WMrlx) ;;;
  Evalue (block, 0%nat).

Definition lock_body : my_op_impl_type op_lock := fun addr =>
  REPEAT CALL C11_cas (addr , 0 , 1 , UMacq , RMrlx) END.

Definition unlock_body : my_op_impl_type op_unlock := fun addr =>
  CALL C11_store (addr , 0 , WMrel).

(** Here is an improved version of acquiring the lock, which waits in a read
  loop when the lock is acquired. *)
Definition lock_alt_body : my_op_impl_type op_lock := fun addr =>
  LOOP 
    LET b <- CALL C11_cas (addr , 0 , 1 , UMacq , RMrlx)  in
    if b then BREAK else 
      REPEAT 
        LET x <- CALL C11_load (addr, RMrlx) in
        Evalue (match x with 0 => true | _ => false end)
      END
  END.
*)
(*

(** Herlihy and Wing queue *)

(* TODO *)
Definition hw_newqueue_body: my_op_impl_type op_newqueue := 
  LET addr <- CALL alloc_op ( ) in 
  Evalue (fst addr).

Definition hw_enqueue_body: my_op_impl_type op_enqueue := 
  fun q v =>
  LET i <- CALL C11_fadd ((q,0), 1, UMrlx) in
  CALL C11_store ((q,i+1), v, WMrel).

Definition hw_dequeue_body : my_op_impl_type op_dequeue := 
  fun q =>
    LOOP 
      LET range <- CALL C11_load ((q,0), RMrlx) in
      FOR i <- 1 TO range DO 
        LET x <- CALL C11_xchg ((q,i), 0, UMacq) in 
        if x =? 0 then Evalue tt else BREAK_RET2 (Some x) 
      END
    END.
*)