Require Import Hahn Operations.
Require Import Setoid Morphisms.

Set Implicit Arguments.

(*****************************************************)

(** Library execution *)

Record tuple := mk_tuple 
  { tup_ev : event -> Prop;
    tup_po : relation event;
    tup_rf : relation event;
    tup_so : relation event;
    tup_hb : relation event }.

Definition tuple_restrict t es so := 
  {| tup_ev := tup_ev t ∩₁ es;
     tup_po := restr_rel es (tup_po t);
     tup_rf := restr_rel es (tup_rf t);
     tup_so := so ;
     tup_hb := restr_rel es (tup_hb t) |}.

Definition tuple_sane t := 
     ⟪ PO_OK: tup_po t ⊆ tup_ev t × tup_ev t ⟫
  /\ ⟪ RF_OK: tup_rf t ⊆ tup_ev t × tup_ev t ⟫
  /\ ⟪ SO_OK: tup_so t ⊆ tup_ev t × tup_ev t ⟫
  /\ ⟪ HB_OK: tup_hb t ⊆ tup_ev t × tup_ev t ⟫.

Record libspec :=
  { libspec_ops : list operation ;
    libspec_logaddr : Type ;
    libspec_requires : event -> libspec_logaddr -> Prop ;
    libspec_allocs : event -> libspec_logaddr -> Prop ;
    libspec_wf : tuple -> Prop ; 
    libspec_consistent : tuple -> Prop ;
    libspec_wf_loc : 
      forall t (WF: libspec_wf t) e (INe : tup_ev t e)
             loc (REQ: libspec_requires e loc),
        exists a,
          ⟪Aa: libspec_allocs a loc⟫ /\ ⟪HBae: (tup_po t ∪ tup_rf t)⁺ a e⟫ ;
    libspec_wf_prefix : 
      forall t (WF: libspec_wf t) (es : event -> Prop)
             (PREFIX: dom_rel ((tup_po t ∪ tup_rf t)⁺ ⨾ ⦗es⦘) ⊆₁ es),
        libspec_wf (tuple_restrict t es (tup_so t)) ;
    libspec_wf_allocs : 
      forall t (WF: libspec_wf t) e (INe : tup_ev t e) e' (INe' : tup_ev t e') (NEQ: e <> e'), 
        set_disjoint (libspec_allocs e) (libspec_allocs e') ;
    libspec_cons_allocs : 
      forall t (CONS: libspec_consistent t) e (INe : tup_ev t e) 
             e' (INe' : tup_ev t e') (NEQ: e <> e'), 
        set_disjoint (libspec_allocs e) (libspec_allocs e') ;
    libspec_cons_sodom : 
      forall t (CONS: libspec_consistent t), tup_so t ⊆ (tup_po t ∪ tup_rf t)⁺ ;
    libspec_equiv:
      forall t (CONS: libspec_consistent t) t'
             (evEQ: tup_ev t ≡₁ tup_ev t')
             (poEQ: tup_po t ≡ tup_po t')
             (rfEQ: tup_rf t ≡ tup_rf t')
             (soEQ: tup_so t ≡ tup_so t')
             (hbEQ: tup_hb t ≡ tup_hb t'),
        libspec_consistent t' ;
    libspec_monotonicity:
      forall t (CONS: libspec_consistent t) hb' 
             (INC1 : (tup_po t ∪ tup_so t)⁺ ⊆ hb') 
             (INC2 : hb' ⊆ tup_hb t),
        libspec_consistent 
          {| tup_ev := tup_ev t;
             tup_po := tup_po t;
             tup_rf := tup_rf t;
             tup_so := tup_so t;
             tup_hb := hb' |} ;
    libspec_locs_ok:
      forall t (SANE: tuple_sane t) (C: libspec_consistent t) (es : event -> Prop) 
             (usedLocs : libspec_logaddr -> Prop)
             (REQ : forall e (INe: tup_ev t e) (REQ_NEMP : ~ libspec_requires e ≡₁ ∅), 
                 es e <-> set_disjoint (libspec_requires e) usedLocs)
             (WFLOC : forall e (INe : tup_ev t e) (INe : es e)
                             loc (REQ: libspec_requires e loc),
                 exists a,
                   ⟪Aa: libspec_allocs a loc⟫ /\ ⟪Ga: tup_ev t a⟫ /\ ⟪RESTRa: es a⟫),
        exists so_restr,
         ⟪SOR: so_restr ⊆ restr_rel es (tup_so t)⟫ /\ 
         ⟪CONSR: libspec_consistent (tuple_restrict t es so_restr)⟫
  }.

Program Example libspec_empty := 
  {| libspec_ops := nil ;
     libspec_logaddr := unit ; 
     libspec_requires e := ∅ ;
     libspec_allocs e := ∅ ;
     libspec_wf := ∅ ; 
     libspec_consistent := ∅ |}.
Solve Obligations with done.


Definition lib_events l e := In (ev_op e) (libspec_ops l).

Definition libs_compatible (libs : libspec -> Prop) := 
  forall l1 (L1: libs l1) l2 (L2: libs l2) op 
         (IN1: lib_events l1 op)
         (IN2: lib_events l2 op), l1 = l2.


(** Executions *)

(* Remaining: define executions and consistent executions. *)
Record execution := mk_execution {
  exec_ev : event -> Prop;
  exec_po : relation event;
  exec_rf : relation event;
  exec_so : relation event
}.

Definition happens_before g := (exec_po g ∪ exec_so g)⁺.

Definition exec_basic_wf g libs := 
  ⟪EV: exec_ev g ⊆₁ set_bunion libs lib_events⟫ /\
  ⟪PO: exec_po g ⊆ exec_ev g × exec_ev g⟫ /\
  ⟪RF: exec_rf g ⊆ ⋃l ∈ libs, (lib_events l ∩₁ exec_ev g) ×
                              (lib_events l ∩₁ exec_ev g)⟫ /\
  ⟪SO: exec_so g ⊆ ⋃l ∈ libs, (lib_events l ∩₁ exec_ev g) ×
                              (lib_events l ∩₁ exec_ev g)⟫ /\
(*  ⟪SOD: exec_so g ⊆ (exec_po g ∪ exec_rf g)⁺⟫ /\ *)
  ⟪IRR: irreflexive (exec_po g ;; (exec_rf g)^*)⁺⟫ /\
  ⟪FS: fsupp (exec_po g ∪ exec_rf g)⁺ ⟫ /\
  ⟪ALLOC: forall lib (ISLIB: libs lib) 
                 a (Ga : exec_ev g a) (LIBa: lib_events lib a)
                 b (Gb : exec_ev g b) (LIBb: lib_events lib b)
                 loc (ALLa : libspec_allocs lib a loc) 
                 (ALLb : libspec_allocs lib b loc),
      a = b⟫.

Definition lib_encapsulated lib (g : event -> Prop) hb :=
  forall e (Ge: g e) (Le: lib_events lib e) 
         loc (Ue: libspec_requires lib e loc),
  exists a, 
    ⟪Aa: libspec_allocs lib a loc⟫ /\
    ⟪Ga: g a⟫ /\ ⟪La: lib_events lib a⟫ /\ 
    ⟪HBae: hb a e⟫.

Definition exec_encapsulated g libs :=
  forall lib (LIB : libs lib) e (Ge: exec_ev g e) (Le: lib_events lib e) 
         loc (Ue: libspec_requires lib e loc),
  exists a, 
    ⟪Aa: libspec_allocs lib a loc⟫ /\
    ⟪Ga: exec_ev g a⟫ /\ ⟪La: lib_events lib a⟫ /\ 
    ⟪HBae: (exec_po g ∪ exec_so g)⁺ a e⟫.

Definition exec_consistent g (libs : libspec -> Prop) := 
  ⟪BWF: exec_basic_wf g libs⟫ /\
  ⟪LIBOK: forall lib (ISLIB: libs lib),
       lib.(libspec_consistent)
         {| tup_ev := exec_ev g ∩₁ lib_events lib ;
            tup_po := restr_rel (lib_events lib) (exec_po g) ;
            tup_rf := restr_rel (lib_events lib) (exec_rf g) ;
            tup_so := restr_rel (lib_events lib) (exec_so g) ;
            tup_hb := restr_rel (lib_events lib) (happens_before g) |} ⟫.

