Require Import Coq.Strings.String.
Require Import Hahn Operations LibBase.
Require Import ZArith Omega.

Set Implicit Arguments.

Definition read_mode_of_rmw m :=
  match m with 
    | UMrlx | UMrel => RMrlx
    | UMacq | UMacqrel => RMacq
    | UMsc => RMsc
  end.

Definition write_mode_of_rmw m :=
  match m with 
    | UMrlx | UMacq => WMrlx
    | UMrel | UMacqrel => WMrel
    | UMsc => WMsc
  end.

Arguments eq_rect [A x P] f [y] e.

(******************************************************************************************)
(******************************************************************************************)
(******************************************************************************************)
Module RC11.
  (* The goal is to define the C11 library and prove it is indeed a library. *)

  Definition is_alloc_ x e := 
    exists (OP: e.(ev_op) = alloc_block), 
      eq_rect e.(ev_res) OP = x.

  Definition is_load_ x v mo e := 
    exists (OP: e.(ev_op) = C11_load),
      eq_rect e.(ev_args) (f_equal _ OP) = (x,(mo,tt)) /\
      eq_rect e.(ev_res) OP = v.

  Definition is_cas_ x v v' mo mo' r e := 
    exists (OP: e.(ev_op) = C11_cas),
      eq_rect e.(ev_args) (f_equal _ OP) = (x,(v,(v',(mo,(mo',tt))))) /\
      eq_rect e.(ev_res) OP = r.

  Definition is_store_ x v mo e := 
    exists (OP: e.(ev_op) = C11_store),
      eq_rect e.(ev_args) (f_equal _ OP) = (x,(v,(mo,tt))).
    
  Definition is_load e := 
    e.(ev_op) = C11_load \/
    exists (OP: e.(ev_op) = C11_cas), 
      eq_rect e.(ev_res) OP <> fst (snd (eq_rect e.(ev_args) (f_equal _ OP))).

  Definition is_alloc e := exists x, is_alloc_ x e.

  Definition is_store e := 
    e.(ev_op) = C11_store.

  Definition is_rmw e := 
    exists (OP: e.(ev_op) = C11_cas), 
      eq_rect e.(ev_res) OP = fst (snd (eq_rect e.(ev_args) (f_equal _ OP))).

  Definition is_fence e := 
    e.(ev_op) = C11_fence.

  Definition has_loc x e := 
    (exists (OP: e.(ev_op) = C11_load),
      fst (eq_rect e.(ev_args) (f_equal _ OP)) = x) \/
    (exists (OP: e.(ev_op) = C11_store),
      fst (eq_rect e.(ev_args) (f_equal _ OP)) = x) \/
    (exists (OP: e.(ev_op) = C11_cas),
      fst (eq_rect e.(ev_args) (f_equal _ OP)) = x).

  Definition has_rval v e := 
    (exists (OP: e.(ev_op) = C11_load), eq_rect e.(ev_res) OP = v) \/
    (exists (OP: e.(ev_op) = C11_cas), eq_rect e.(ev_res) OP = v).

  Definition has_wval v e := 
    (exists (OP: e.(ev_op) = C11_store),
      fst (snd (eq_rect e.(ev_args) (f_equal _ OP))) = v) \/
    (exists (OP: e.(ev_op) = C11_cas),
      eq_rect e.(ev_res) OP = fst (snd (eq_rect e.(ev_args) (f_equal _ OP))) /\
      fst (snd (snd (eq_rect e.(ev_args) (f_equal _ OP)))) = v).

  Definition is_read e := is_load e \/ is_rmw e.
  Definition is_write e := is_store e \/ is_rmw e.
  Definition is_access e := is_load e \/ is_store e \/ is_rmw e.

  Lemma is_read_access e (H: is_read e) : is_access e.
  Proof. red; red in H; tauto. Qed.

  Lemma is_write_access e (H: is_write e) : is_access e.
  Proof. red; red in H; tauto. Qed.

  Lemma is_accessErw e : is_access e <-> is_read e \/ is_write e.
  Proof. unfold is_access, is_read, is_write; tauto. Qed.

  Hint Resolve is_read_access is_write_access.

  Lemma has_loc_is_access e x : has_loc x e -> is_access e.
  Proof. 
    unfold has_loc, is_access, is_load, is_store, is_rmw; ins; desf; eauto. 
    tertium_non_datur (eq_rect (ev_res e) OP = fst (snd (eq_rect (ev_args e) (f_equal _ OP)))); desf; eauto.
  Qed.

  Lemma has_rval_is_read e v : has_rval v e -> is_read e.
  Proof. 
    unfold has_rval, is_read, is_load, is_rmw; ins; desf; eauto.
    tertium_non_datur (eq_rect (ev_res e) OP = fst (snd (eq_rect (ev_args e) (f_equal _ OP)))); desf; eauto.
  Qed.

  Lemma has_wval_is_write e v : has_wval v e -> is_write e.
  Proof. unfold has_wval, is_write, is_store, is_rmw; ins; desf; eauto. Qed.

  Hint Resolve has_loc_is_access has_rval_is_read has_wval_is_write.

  Lemma is_access_has_loc e : is_access e -> exists x, has_loc x e.
  Proof. 
    unfold has_loc, is_access, is_load, is_store, is_rmw; ins; desf; 
      destruct e; ins; subst; 
    eexists; first [left; exists eq_refl | 
                    right; left; exists eq_refl | 
                    right; right; exists eq_refl]; auto.
  Qed.

  Lemma has_loc_unique x x' e : has_loc x e -> has_loc x' e -> x = x'.
  Proof.
    by unfold has_loc; destruct e; ins; desf; rewrite <- !eq_rect_eq.
  Qed.  

  Lemma is_read_has_rval e : is_read e -> exists v, has_rval v e.
  Proof. 
    unfold has_rval, is_read, is_load, is_rmw; ins; desf; destruct e; ins; subst; ins. 
    all: eexists; constructor; exists eq_refl; ins.
  Qed.

  Lemma is_write_has_wval e : is_write e -> exists v, has_wval v e.
  Proof. 
    unfold has_wval, is_write, is_store, is_rmw; ins; desf; destruct e; ins; subst; ins.
    all: eexists; constructor; exists eq_refl; ins.
  Qed.

  Lemma is_readE e : is_read e <-> exists v, has_rval v e.
  Proof. 
    split; ins; desf; eauto using is_read_has_rval, has_rval_is_read.
  Qed.

  Lemma is_writeE e : is_write e <-> exists v, has_wval v e.
  Proof. 
    split; ins; desf; eauto using is_write_has_wval, has_wval_is_write. 
  Qed.

  Lemma is_accessE e : is_access e <-> exists x, has_loc x e.
  Proof. 
    split; ins; desf; eauto using is_access_has_loc, has_loc_is_access. 
  Qed.
 
  Lemma functional_has_wval : 
    forall e v1 (V: has_wval v1 e) v2 (V': has_wval v2 e), v1 = v2.
  Proof.
    unfold has_wval; ins; desf; simpls; destruct e; ins; desf;
    rewrite <- !eq_rect_eq; ins.
  Qed.

  Lemma functional_has_rval : 
    forall e v1 (V: has_rval v1 e) v2 (V': has_rval v2 e), v1 = v2.
  Proof.
    unfold has_rval; ins; desf; simpls; destruct e; ins; desf;
    rewrite <- !eq_rect_eq; ins.
  Qed.

  Definition has_read_mode m e := 
    (exists (OP: e.(ev_op) = C11_load),
      fst (snd (eq_rect e.(ev_args) (f_equal _ OP))) = m) \/
    (exists (OP: e.(ev_op) = C11_cas),
      eq_rect e.(ev_res) OP <> fst (snd (eq_rect e.(ev_args) (f_equal _ OP))) /\
      fst (snd (snd (snd (snd (eq_rect e.(ev_args) (f_equal _ OP)))))) = m) \/
    (exists (OP: e.(ev_op) = C11_cas),
      eq_rect e.(ev_res) OP = fst (snd (eq_rect e.(ev_args) (f_equal _ OP))) /\
      read_mode_of_rmw 
        (fst (snd (snd (snd (eq_rect e.(ev_args) (f_equal _ OP)))))) = m) \/
    (exists (OP: e.(ev_op) = C11_fence),
      read_mode_of_rmw (fst (eq_rect e.(ev_args) (f_equal _ OP))) = m).

  Definition has_write_mode m e := 
    (exists (OP: e.(ev_op) = C11_store),
      fst (snd (snd (eq_rect e.(ev_args) (f_equal _ OP)))) = m) \/
    (exists (OP: e.(ev_op) = C11_cas),
      eq_rect e.(ev_res) OP = fst (snd (eq_rect e.(ev_args) (f_equal _ OP)))  /\
      write_mode_of_rmw 
        (fst (snd (snd (snd (eq_rect e.(ev_args) (f_equal _ OP)))))) = m) \/
    (exists (OP: e.(ev_op) = C11_fence),
      write_mode_of_rmw (fst (eq_rect e.(ev_args) (f_equal _ OP))) = m).

  Definition is_acquire e := 
    exists m, 
      has_read_mode m e /\ 
      match m with 
        RMna | RMrlx => False
      | RMacq | RMsc => True
      end.

  Definition is_release e := 
    exists m, 
      has_write_mode m e /\ 
      match m with 
        WMna | WMrlx => False
      | WMrel | WMsc => True
      end.

  Definition eco (rf mo : relation event) :=  ((rf⁻¹)^? ⨾ mo \ ⦗fun _ => True⦘) ⨾ rf^? ∪ rf.

  Definition wf (t : tuple) := 
      << EOK: tup_ev t ⊆₁ is_alloc ∪₁ is_access ∪₁ is_fence >>
   /\ << WF_LOC: 
        forall x, tup_ev t ∩₁ has_loc x 
                ⊆₁ codom_rel (⦗is_alloc_ (fst x)⦘ ⨾ (tup_po t ∪ tup_rf t)⁺) >>
   /\ << ALLOC : forall x a (EVa: tup_ev t a) (Aa: is_alloc_ x a)
                        b (EVb: tup_ev t b) (Ab: is_alloc_ x b), a = b >>.

  Definition consistent t :=
      exists mo, 
       ⟪ EOK: tup_ev t ⊆₁ is_alloc ∪₁ is_access ∪₁ is_fence ⟫ 
      /\ ⟪ ALLOC : forall x a (EVa: tup_ev t a) (Aa: is_alloc_ x a)
                           b (EVb: tup_ev t b) (Ab: is_alloc_ x b), a = b ⟫
      /\ ⟪ RFOK: tup_rf t ⊆ is_write × is_read ⟫
      /\ ⟪ RFVAL: tup_rf t ⊆ ⋃ v, has_wval v × has_rval v ⟫
      /\ ⟪ RFLOC: tup_rf t ⊆ ⋃ x, has_loc x × has_loc x ⟫
      /\ ⟪ CODOM: codom_rel (tup_rf t) ≡₁ is_read ∩₁ tup_ev t ⟫
      /\ ⟪ FUNC: functional (tup_rf t)⁻¹ ⟫
      /\ ⟪ SOEQ: tup_so t ≡ ⦗is_release⦘ ⨾ (⦗is_fence⦘ ⨾ tup_po t)^? 
                     ⨾ (tup_rf t ⨾ ⦗is_rmw⦘)^* ⨾ tup_rf t 
                     ⨾ (tup_po t ⨾ ⦗is_fence⦘)^? ⨾ ⦗is_acquire⦘ ⟫
      /\ ⟪ MO: strict_partial_order mo ⟫
      /\ ⟪ MOT: forall l, is_total (is_write ∩₁ has_loc l ∩₁ tup_ev t) mo ⟫
      /\ ⟪ ACYC: acyclic (tup_hb t ∪ mo ∪ tup_rf t ∪ 
                           ((tup_rf t)⁻¹ ⨾ mo) \ ⦗fun _ => True⦘) ⟫.

  Local Ltac u := 
    unfold consistent, wf, is_alloc, is_access, is_fence, is_read, is_write, is_store, is_load, 
      is_rmw, is_alloc_, is_store_, is_load_, has_loc in *;
    ins; desf.

  Lemma acyclic_rf t : consistent t -> acyclic (tup_rf t).
  Proof.     
    intros; repeat red in H; desf. 
    eapply acyclic_mon; eauto with hahn. 
  Qed.

  Lemma so_in_porf t (H: consistent t): tup_so t ⊆ (tup_po t ∪ tup_rf t)^+.
  Proof. 
    cdes H; rewrite SOEQ.
    rewrite !inclusion_seq_eqv_l, !inclusion_seq_eqv_r; rels. 
    arewrite (tup_rf t ⊆ tup_po t ∪ tup_rf t).
    arewrite (tup_po t ⊆ tup_po t ∪ tup_rf t); rels.
    eapply inclusion_t_r_t; autos.
  Qed.

  Program Definition library := 
    {| libspec_ops := alloc_block :: C11_load :: C11_store :: C11_cas :: nil ; 
       libspec_logaddr := address ;
       libspec_allocs e x := is_alloc_ (fst x) e ; 
       libspec_requires e x := has_loc x e ;
       libspec_wf := wf ;
       libspec_consistent := consistent |}.
  Next Obligation.
    u; unfolder in *; specialize_full EOK; eauto; destruct e; ins; desf; ins; desf.
    all: forward eapply WF_LOC as X; [split; try edone; ins;
           try first [left; exists eq_refl | right; constructor; exists eq_refl]; ins|desf].
    all: exists z; splits; eauto; destruct z; ins; desf; ins; desf; ins.
  Qed.
  Next Obligation.
    u; desf; splits; ins.
      by rewrite <- EOK; ins; unfolder; ins; desf. 
    2: by unfold set_inter in *; desc; eapply ALLOC with (x:=x); eauto.
    rewrite set_interC with (s' := es), set_interA, WF_LOC; clear WF_LOC.
    unfolder in *; ins; desf; eexists; split; eauto; clear OP H0.
    induction H1; [eapply t_step|eapply t_trans]; desf; eauto 10 using t_step, t_trans.
  Qed.
  Next Obligation.
    unfold wf in *; desf; splits; ins.
    red; ins; destruct NEQ; eapply ALLOC with (x:=fst x); eauto. 
  Qed.
  Next Obligation.
    unfold consistent in *; desf; splits; ins.
    red; ins; destruct NEQ; eapply ALLOC with (x:=fst x); eauto. 
  Qed.
  Next Obligation.
    auto using so_in_porf.
  Qed.
  Next Obligation.
    unfold consistent, strict_partial_order in *; ins; desf. 
    exists mo; splits; ins; eauto. 
    2: by unfolder in *; ins; desf; eauto.
    all: by rewrite <- ?evEQ, <- ?poEQ, <- ?rfEQ, <- ?soEQ, <- ?hbEQ.
  Qed.
  Next Obligation.
    unfold consistent in *; ins; desf; splits; eauto. 
    exists mo; splits; eauto; rewrite INC2; eauto.
  Qed.
  Next Obligation.
    unfold tuple_sane, consistent, strict_partial_order in *; ins; desf.
    assert (RFEQ: restr_rel es (tup_rf t) ≡ tup_rf t ⨾ ⦗es⦘).
    {
      clear -REQ RFLOC RF_OK; split; unfolder; ins; desf; splits; ins.
      unfolder in RF_OK; specialize (RF_OK _ _ H); desc.
      specialize_full RFLOC; eauto; desf; unfold has_loc in *; desf.
      unfolder in RFLOC; specialize (RFLOC _ _ H); desf.
      all: rewrite REQ; ins; [rewrite REQ in H0; ins|]; clear REQ.
      all: destruct x, y; ins; desf; ins; desf; ins.
      all: try by apply set_nonemptyE; repeat eexists; repeat constructor; exists eq_refl; ins. 
      all: unfolder in *; ins; desf; ins; rewrite <- eq_rect_eq in *; desf.
      all: eapply H0 in IN'; eauto.
      all: by repeat eexists; repeat constructor; eexists eq_refl; ins. 
    }
    eexists; splits; cycle 1.
    exists (restr_rel es mo); splits; ins; 
    try solve [rewrite ?inclusion_restr; eauto with hahn].
    - rewrite EOK; unfolder; tauto.
    - by unfolder in *; ins; desf; eauto.
    - by rewrite RFEQ, <- set_interA, <- CODOM; clear; unfolder; intuition; desf; eauto.
    - by unfolder; ins; desf; eauto. 
    - by unfolder; ins; desf; eauto; eapply MOT in NEQ; unfolder; desf; eauto.

    - rewrite SOEQ, !restr_relE, !seqA; 
      arewrite (⦗es⦘ ⨾ (⦗es⦘ ⨾ tup_po t ⨾ ⦗es⦘ ⨾ ⦗is_fence⦘)^? ⊆ (tup_po t ⨾ ⦗is_fence⦘)^? ⨾ ⦗es⦘).
        by unfolder; ins; desf; eauto 6.
      rewrite seq_eqvC with (doma := is_acquire); hahn_frame. 
      arewrite ((⦗es⦘ ⨾ tup_rf t ⨾ ⦗es⦘ ⨾ ⦗is_rmw⦘)＊ ⨾ ⦗es⦘ ⊆ ⦗es⦘ ⨾ (tup_rf t ⨾ ⦗is_rmw⦘)＊).
        rewrite rt_begin, !seq_union_l, seq_id_l, !seqA.
        apply inclusion_union_l; hahn_frame. 
        rewrite ?inclusion_seq_eqv_r with (dom := es), ?inclusion_seq_eqv_l with (dom := es); ins.
        rewrite <- seqA, <- ct_begin; eauto with hahn.
      seq_rewrite (seq_eqvC es is_release); hahn_frame_l; hahn_frame_r.
      by unfolder; ins; desf; eauto 6.
  Qed.

End RC11.

