Require Import Coq.Strings.String.
Require Import Hahn Omega Operations LibBase.
Require Import Wf IndefiniteDescription.

Set Implicit Arguments.

Module LockLib.

  Definition is_lock e := e.(ev_op) = op_lock.
  Definition is_unlock e := e.(ev_op) = op_unlock. 

  Definition is_newlock_ok l e := 
    exists (OP: e.(ev_op) = op_newlock), 
      (eq_rect _ _ e.(ev_res) _ OP) = l.

  Definition is_initlock_ok l e := 
    exists (OP: e.(ev_op) = op_initlock), 
      fst (eq_rect _ _ e.(ev_args) _ (f_equal _ OP)) = l.

  Definition is_lock_ok l e := 
    exists (OP: e.(ev_op) = op_lock), 
      fst (eq_rect _ _ e.(ev_args) _ (f_equal _ OP)) = l.

  Definition is_unlock_ok l e := 
    exists (OP: e.(ev_op) = op_unlock), 
      fst (eq_rect _ _ e.(ev_args) _ (f_equal _ OP)) = l.

  Definition is_ok l := 
    is_newlock_ok l ∪₁ is_lock_ok l ∪₁ is_unlock_ok l.
    
  Definition wf (t : tuple) := 
    exists flock, 
      ⟪WFev: tup_ev t ⊆₁ ⋃₁l, is_ok l⟫ /\
      ⟪WFloc: forall l, tup_ev t ∩₁ (is_lock_ok l ∪₁ is_unlock_ok l) 
                ⊆₁ codom_rel (⦗is_newlock_ok l⦘ ⨾ (tup_po t ∪ tup_rf t)⁺)⟫ /\
      ⟪WFpair: forall x (EV: tup_ev t x) l (UNL: is_unlock_ok l x), is_lock_ok l (flock x)⟫ /\
      ⟪WFpo: forall x (EV: tup_ev t x) (UNL: is_unlock x), t.(tup_po) (flock x) x⟫ /\
      ⟪WFimm: forall l x (UNL: is_unlock_ok l x) y (PO1: t.(tup_po) (flock x) y) 
                       (PO2: t.(tup_po) y x) (L: is_ok l y), False⟫ /\
      ⟪ALLOC: forall l a (EVa: tup_ev t a) (Aa: is_newlock_ok l a)
                   b (EVb: tup_ev t b) (Ab: is_newlock_ok l b), a = b⟫.

  Definition consistent (t : tuple) :=
    ⟪EOK: tup_ev t ⊆₁ ⋃₁l, is_ok l⟫ /\
    ⟪ALLOC: forall l a (EVa: tup_ev t a) (Aa: is_newlock_ok l a)
                   b (EVb: tup_ev t b) (Ab: is_newlock_ok l b), a = b⟫ /\
    ⟪RFDOM: tup_rf t ⊆ ⋃l, (is_newlock_ok l ∪₁ is_unlock_ok l) × is_lock_ok l⟫ /\
    ⟪RFFUN: functional (tup_rf t)⟫ /\
    ⟪RF1FUN: functional (tup_rf t)⁻¹⟫ /\ 
    ⟪SOEQ: tup_so t ≡ tup_rf t⟫ /\ 
    ⟪CODOM: codom_rel (tup_rf t) ≡₁ is_lock ∩₁ tup_ev t⟫.

  Local Ltac u := 
    unfold wf, consistent, is_ok, is_unlock_ok, is_lock_ok, is_newlock_ok in *.

  Lemma unique_loc: forall e l l' (OK: is_ok l e) (OK': is_ok l' e), l = l'.
  Proof.
    by u; unfolder; destruct e; ins; desf; rewrite <- ?eq_rect_eq in *.
  Qed.  

  Lemma is_lock_okE l : is_lock_ok l ≡₁ is_ok l ∩₁ is_lock.
  Proof. 
    u; unfolder; splits; destruct x; splits; ins; desf.
    all: repeat constructor; exists eq_refl; ins.
  Qed.

  Lemma is_unlock_okE l : is_unlock_ok l ≡₁ is_ok l ∩₁ is_unlock.
  Proof. 
    u; unfolder; splits; destruct x; splits; ins; desf.
    all: repeat constructor; exists eq_refl; ins.
  Qed.

  Lemma is_lock_okI l e : is_lock e -> is_ok l e -> is_lock_ok l e.
  Proof. ins; apply is_lock_okE; ins. Qed. 

  Lemma is_unlock_okI l e : is_unlock e -> is_ok l e -> is_unlock_ok l e.
  Proof. ins; apply is_unlock_okE; ins. Qed. 

  Lemma acyclic_rf t : consistent t -> acyclic (tup_rf t).
  Proof.     
    intros; repeat red in H; desf. 
    eapply acyclic_mon, RFDOM.
    clear; unfolder; repeat red.
    unfold is_newlock_ok, is_unlock_ok, is_lock_ok.
    ins; desf; apply ct_begin in H; unfolder in H; ins; desc.
    apply rt_begin in H0; unfolder in H0; ins; desf.
    all: destruct x; ins; desf; destruct z; ins; desf.
  Qed.

  Program Definition library := 
    {| libspec_ops := op_newlock :: op_lock :: op_unlock :: nil ;
       libspec_logaddr := address ;
       libspec_allocs e x := is_newlock_ok x e ; 
       libspec_requires e x := is_lock_ok x e \/ is_unlock_ok x e ;
       libspec_wf := wf ;
       libspec_consistent := consistent |}.
  Next Obligation.
    u; unfolder in *; ins; desf. 
    all: specialize_full WFev; eauto; destruct e; ins; desf; ins; desf; ins.
    all: forward eapply WFloc as X; [split; try edone; ins; constructor; exists eq_refl; ins|desf; eauto]. 
  Qed.
  Next Obligation.
    unfold wf in *; ins; desf; exists flock; splits; ins.
    all: unfolder in *; ins; desc; splits; eauto 8 using t_step. 
    specialize_full WFloc; eauto; clear H H0; desc.
    eexists; split; eauto; clear WFloc1; desf.
    induction WFloc0; [eapply t_step|eapply t_trans]; desf; eauto 10 using t_step, t_trans.
  Qed.
  Next Obligation.
    unfold wf in *; desf; splits; ins.
    red; ins; destruct NEQ; eapply ALLOC with (l:=x); eauto.
  Qed.
  Next Obligation.
    unfold consistent in *; desf; splits; ins.
    red; ins; destruct NEQ; eapply ALLOC with (l:=x); eauto.
  Qed.
  Next Obligation.
    unfold consistent in *; ins; desf. 
    rewrite SOEQ; unfolder; eauto 10 using t_step.
  Qed.
  Next Obligation.
    unfold consistent in *; ins; desf. 
    splits; ins; eauto. 
    2: by unfolder in *; ins; desc; eauto.
    all: by rewrite <- ?evEQ, <- ?poEQ, <- ?rfEQ, <- ?soEQ, <- ?hbEQ.
  Qed.
  Next Obligation.
    unfold tuple_sane, consistent in *; ins; desf.
    eexists; splits; ins; eauto; try (by rewrite SOEQ).
      by unfolder in *; ins; desc; eauto.
      by unfolder in *; ins; desc; eauto.
    all: try solve [rewrite inclusion_restr; ins].
    rewrite <- set_interA, <- CODOM; unfolder; split; ins; desf; eauto.
    rename H into RF.
    enough (es x0) by eauto 8.
    specialize (RF_OK _ _ RF); unfolder in RF_OK; desc.
    apply RFDOM in RF; clear RFDOM; unfolder in RF; desf.
      by eapply WFLOC in H0; eauto; desf; enough (a0 = x0); desf; eauto.
    eapply REQ in H0; ins; try eapply REQ; ins; clear REQ; try apply set_nonemptyE; eauto.
    red; ins; eapply H0 in IN'; eauto.
    enough (x1 = a); subst; eauto.
    eapply unique_loc with (e:=x0); red; unfolder; tauto. 
  Qed.

End LockLib.

Hint Immediate LockLib.is_lock_okI LockLib.is_unlock_okI.
