Require Import Coq.Strings.String.
Require Import Hahn Omega LibBase Operations.
Require Import Wf IndefiniteDescription.
Require Import Sorted.
Require Import Setoid.
Require Import ProofIrrelevance.

Set Implicit Arguments.

Module QueueLib.
  Definition is_newqueue_loc l e := 
    exists (OP: e.(ev_op) = op_newqueue), 
      (eq_rect _ _ e.(ev_res) _ OP) = l.

  Definition is_newqueue e := e.(ev_op) = op_newqueue.
  Definition is_enqueue e := e.(ev_op) = op_enqueue.
  Definition is_dequeue e := e.(ev_op) = op_dequeue.

  Definition is_enqueue_loc l e := 
    exists (OP: e.(ev_op) = op_enqueue), 
      fst (eq_rect _ _ e.(ev_args) _ (f_equal _ OP)) = l.

  Definition is_enqueueV' v e := 
    exists (OP: e.(ev_op) = op_enqueue), 
      fst (snd (eq_rect _ _ e.(ev_args) _ (f_equal _ OP))) = v.

  Definition is_enqueueV l v e := 
    exists (OP: e.(ev_op) = op_enqueue), 
      fst (eq_rect _ _ e.(ev_args) _ (f_equal _ OP)) = l /\
      fst (snd (eq_rect _ _ e.(ev_args) _ (f_equal _ OP))) = v.

  Definition is_dequeue_loc l e := 
    exists (OP: e.(ev_op) = op_dequeue), 
      fst (eq_rect _ _ e.(ev_args) _ (f_equal _ OP)) = l.
  
  Definition is_dequeueV' v e :=
    exists (OP: e.(ev_op) = op_dequeue), 
      match (eq_rect _ (fun x => x) e.(ev_res) _ (f_equal _ OP)) with
        | Some v' => v = v'
        | _ => False 
      end.
      
  Definition is_dequeueV l v e :=
    exists (OP: e.(ev_op) = op_dequeue), 
      fst (eq_rect _ _ e.(ev_args) _ (f_equal _ OP)) = l /\
      match (eq_rect _ (fun x => x) e.(ev_res) _ (f_equal _ OP)) with
        | Some v' => v = v'
        | _ => False 
      end.

  Definition is_dequeueN l e :=
    exists (OP: e.(ev_op) = op_dequeue), 
      fst (eq_rect _ _ e.(ev_args) _ (f_equal _ OP)) = l /\
      match (eq_rect _ (fun x => x) e.(ev_res) _ (f_equal _ OP)) with
        | None => True
        | _ => False
      end.

  Lemma is_dequeueV_weaken l v e: is_dequeueV l v e -> is_dequeueV' v e.
  Proof. 
    unfold is_dequeueV, is_dequeueV'; destruct e; ins; desf.
    exists eq_refl; rewrite <- eq_rect_eq in *; desf.
  Qed.

  (* Need this to state basic wf conditions. *)
  Definition has_loc x := is_enqueue_loc x ∪₁ is_dequeue_loc x.
    
  Lemma is_enqueue_impl l v e : is_enqueueV l v e -> is_enqueue_loc l e.
  Proof.
    ins. red in H; red.
    des; exists OP.
    eauto.
  Qed.
  Lemma is_dequeue_impl l v e : is_dequeueV l v e -> is_dequeue_loc l e.
  Proof.
    ins. red in H; red.
    des; exists OP.
    eauto.
  Qed.

  Lemma is_dequeueN_impl l e : is_dequeueN l e -> is_dequeue_loc l e.
  Proof.
    ins. red in H; red.
    des; exists OP.
    eauto.
  Qed.

  Lemma newqueue_loc_unique x l1 l2: is_newqueue_loc l1 x /\ is_newqueue_loc l2 x -> l1 = l2.
  Proof.
    ins; des.
    unfold has_loc in *; des.
    red in H; red in H0; des.
      by destruct x; desf; ins; desf; ins; rewrite <- eq_rect_eq in *. 
  Qed.

  Lemma dequeue_loc_unique x l1 l2: is_dequeue_loc l1 x /\ is_dequeue_loc l2 x -> l1 = l2.
  Proof.
    ins; des.
    unfold has_loc in *; des.
    red in H; red in H0; des.
      by destruct x; desf; ins; desf; ins; rewrite <- eq_rect_eq in *. 
  Qed.

  Lemma enqueue_loc_unique x l1 l2: is_enqueue_loc l1 x /\ is_enqueue_loc l2 x -> l1 = l2.
  Proof.
    ins; des.
    red in H; red in H0; des.
      by destruct x; desf; ins; desf; ins; rewrite <- eq_rect_eq in *. 
  Qed.

  Lemma is_dequeue_alt_def l e: is_dequeue_loc l e <-> is_dequeueN l e \/ exists v, is_dequeueV l v e.
  Proof.
    unfold is_dequeue_loc, is_dequeueN, is_dequeueV.
    split; ins; des; ins.
    2,3: by exists OP; ins.
    classical_left; exists OP.
    split; ins.
    case_eq 
      (eq_rect (op_result (ev_op e)) (fun x : Type => x) (ev_res e) (option Z) (f_equal op_result OP)); 
      ins.
    apply H0.
    exists z; exists OP; ins; split; ins.
    rewrite H1; trivial.
  Qed.

  Fixpoint is_fifo com l s S := 
    match S with
    | nil => True
    | ev :: S' => is_enqueue_loc l ev /\ is_fifo com l (s ++ ev :: nil) S' \/
                 match s with 
                 | nil => is_dequeueN l ev /\ is_fifo com l s S'
                 | ev' :: s' => com ev' ev /\ is_fifo com l s' S'
                 end
    end.

  Lemma has_loc_unique: forall l l' e, has_loc l e -> has_loc l' e -> l = l'.
  Proof.
    unfold has_loc; unfolder; ins; desf; eauto using enqueue_loc_unique, dequeue_loc_unique.
    all: destruct e; red in H, H0 ; desf; ins; desf.
  Qed.

  Lemma is_fifo_same_loc: 
    forall com l s S (FIFO: is_fifo com l s S) e (IN: In e (s ++ S))
      (sOK: forall e, In e s -> has_loc l e)
      (COMSUB: com ⊆ ⋃l, ⋃v, is_enqueueV l v × is_dequeueV l v),
    has_loc l e.
  Proof.
    intros until S; revert s.
    induction S; ins.
      by apply sOK; rewrite app_nil_r in IN.
    desf; desf.
    * eapply IHS; eauto; ins.
        by rewrite <- app_assoc, <- app_comm_cons.
      apply in_app_iff in H; ins; desf; eauto; vauto.
    * ins; desf.
      by apply is_dequeueN_impl in FIFO; vauto.
      by eapply IHS; eauto.
    * apply in_app_iff in IN; desf; eauto.
      ins; desf.
      + specialize (sOK _ (or_introl eq_refl)).
        apply COMSUB in FIFO; unfolder in FIFO; desf.
        apply is_enqueue_impl in FIFO2.
        enough (l = a) by (apply is_dequeue_impl in FIFO3; vauto).
        eapply has_loc_unique; eauto; vauto.
      + eapply IHS; eauto.
        apply in_app_iff; eauto.
  Qed.

  Lemma fifo_loc_restr s (com : relation event) l S (COMSUB: com ⊆ ⋃l, ⋃v, is_enqueueV l v × is_dequeueV l v)
        (sOK: forall e, In e s -> has_loc l e):
          is_fifo com l s S <-> is_fifo (<| has_loc l |> ;; com ;; <| has_loc l |>) l s S.
  Proof.
    split.
    - revert sOK; revert s; induction S; ins; des.
      + left; split; ins. 
        apply IHS; ins.
        apply in_app_iff in H1; des.
          by eauto.
          by simpl in H1; des; ins; subst; vauto.
      + right; case_eq s.
          by ins; subst; des; split; ins; apply IHS.
        ins; rewrite H0 in *;
        des; split; [| apply IHS; ins]; remember H as C; clear HeqC.
        assert (QQQ: has_loc l e)
           by (apply sOK; left; ins).
        assert (WWW: has_loc l a).
        {  apply COMSUB in C; right.
           unfolder in C; des.
           apply is_enqueue_impl in C1.
           apply is_dequeue_impl in C2.
           enough (l = a0); subst; ins.
           apply enqueue_loc_unique with (x:=e); split; ins.
           specialize (sOK _ (or_introl eq_refl)).
           do 2 (red in sOK); des; ins.
           destruct e; red in sOK, C1; des; ins; desf. }
         by apply COMSUB in H; exists e; split; try exists a; split.
         by apply sOK; right.
    - revert sOK; revert s; induction S; ins; des.
      + left; split; ins; apply IHS; ins.
        apply in_app_iff in H1; des.
          by eauto.
          by simpl in H1; des; ins; subst; vauto.
      + right; case_eq s.
          by ins; subst; des; split; ins; apply IHS; ins.
         ins; rewrite H0 in *.
         des; split; [| apply IHS; ins]; remember H as C; clear HeqC.
           by unfolder in C; desf.
           by apply sOK; right.
    Qed.

  Add Parametric Morphism: is_fifo with signature
    same_relation ==> eq ==> eq ==> eq ==> iff as is_fifo_mor.
  Proof.
    intros com com' EQ l s S.
    revert s.
    induction S; ins.
    split; ins; desf; ins; desf; ins;
    solve [ by left; splits; eauto; apply IHS 
          | by right; splits; eauto; apply IHS
          | by right; splits; eauto; first [apply EQ | apply IHS]].
  Qed.

  Definition wf (t: tuple) :=
      ⟪EOK: tup_ev t ⊆₁ is_newqueue ∪₁ is_enqueue ∪₁ is_dequeue⟫ /\
      ⟪WF_LOC: forall x, tup_ev t ∩₁ has_loc x
                ⊆₁ codom_rel (⦗is_newqueue_loc x⦘ ⨾ (tup_po t ∪ tup_rf t)⁺)⟫ /\
      ⟪ALLOC : forall x a (EVa: tup_ev t a) (Aa: is_newqueue_loc x a)
                        b (EVb: tup_ev t b) (Ab: is_newqueue_loc x b), a = b⟫.

  Definition finite_prefix A (to : relation A) ls := 
      ⟪RESP: StronglySorted to ls⟫ /\
      ⟪FIN: ls = nil \/ exists el, In el ls /\ forall el', to el' el <-> In el' ls⟫.

  Definition consistent (t : tuple) :=
    ⟪RFOK: tup_rf t ⊆ ⋃l, ⋃v, is_enqueueV l v × is_dequeueV l v⟫ /\
    ⟪ALLOC: forall x a (EVa: tup_ev t a) (Aa: is_newqueue_loc x a)
                           b (EVb: tup_ev t b) (Ab: is_newqueue_loc x b), a = b⟫ /\
    ⟪RFFUN: functional (tup_rf t)⟫ /\
    ⟪RF1FUN: functional (tup_rf t)⁻¹⟫ /\
    ⟪SOEQ: tup_so t ≡ tup_rf t⟫ /\
    ⟪DEQN: forall l, tup_ev t ∩₁ (is_dequeue_loc l \₁ codom_rel (tup_rf t)) ⊆₁ is_dequeueN l⟫ /\
    ⟪SEQEN: forall l, exists to, strict_total_order (tup_ev t ∩₁ has_loc l) to /\
                               restr_rel (has_loc l) (tup_hb t) ⊆ to /\
            forall S (SDOM: (fun e => In e S) ⊆₁ tup_ev t ∩₁ has_loc l) 
                   (SPREFIX:finite_prefix to S),
              is_fifo (tup_rf t) l nil S⟫.

  Lemma acyclic_rf t : consistent t -> acyclic (tup_rf t).
  Proof.     
    intros; repeat red in H; desf. 
    eapply acyclic_mon; eauto with hahn. 
    clear; unfolder; repeat red.
    unfold is_enqueueV, is_dequeueV.
    ins; desf; apply ct_begin in H; unfolder in H; ins; desc.
    apply rt_begin in H0; unfolder in H0; ins. 
    desf; destruct x; ins; desf; destruct z; ins; desf.
  Qed.

  Lemma functional_subs A (r r' : relation A) (SUBS: r ⊆ r') (FUNC: functional r') :
    functional r.
  Proof. unfold functional in *; eauto. Qed.

  Lemma restr_subs A (r : relation A) (s : A -> Prop): 
              ⦗s⦘ ⨾ r ⨾ ⦗s⦘ ⊆ r.
  Proof. unfolder; splits; des; ins. Qed.

  Lemma StronglySorted_empty_rel: forall A S, @StronglySorted A ∅₂ S <-> length S < 2.
  Proof.
    split; ins.
    * induction H; ins; try omega.
      induction H0; ins; omega.
    * destruct S; ins; try by constructor.
      destruct S; ins; try omega.
      vauto.
  Qed.

  Ltac falsify_ := 
    match goal with 
    | H1: is_enqueue_loc _ ?x, H2: is_dequeue_loc _ ?x |- _ => destruct x; red in H1, H2 ; desf; ins; desf
    end.

  Ltac eqloc_ :=
    match goal with
    | H1: is_enqueue_loc ?a ?x, H2: is_enqueue_loc ?b ?x |- ?a = ?b
            => eapply enqueue_loc_unique; ins; eauto 
    | H1: is_dequeue_loc ?a ?x, H2: is_dequeue_loc ?b ?x |- ?a = ?b
            => eapply dequeue_loc_unique; ins; eauto
    end; subst.

  Program Definition library := 
    {| libspec_ops := op_newqueue :: op_enqueue :: op_dequeue :: nil ;
       libspec_logaddr := queue_address ;
       libspec_allocs e x := is_newqueue_loc x e ;
       libspec_requires e x := has_loc x e ;
       libspec_wf := wf ;
       libspec_consistent := consistent |}.
  Next Obligation.
    unfold wf in *; des;
    assert(AA: (tup_ev t ∩₁ has_loc loc) e) by firstorder;
    apply WF_LOC in AA;
    red in AA; des; exists x;
    firstorder; subst; ins.
  Qed.
  Next Obligation.
    unfold wf in *; splits; des; ins.
      by intros x H; red in H; des; apply EOK.
    - rewrite union_restr.
      rewrite (set_interC _ es), set_interA, WF_LOC.
      revert PREFIX; clear.
      generalize (tup_po t ∪ tup_rf t) as r; ins; unfolder in *; ins; desf.
      eexists; split; eauto.
      clear H0; induction H1; eauto 10 using clos_trans.
   - red in EVa, EVb; des; eapply ALLOC; eauto.
  Qed.
  Next Obligation.
    unfold wf in *; des; red; eauto.
  Qed.
  Next Obligation.
    unfold consistent in *; des; red; eauto.
  Qed.
  Next Obligation.
    unfold consistent in *; des; rewrite SOEQ; unfolder; eauto 10 using t_step.
  Qed.
  Next Obligation.
    unfold consistent in *; des; splits; ins.
    all: try repeat match goal with
    | [ H: same_relation _ _ |- _ ] => ins; rewrite <- H; eauto
    | [ H: forall l:queue_address, _ |- _ ] => specialize (H l)
    end.
      by apply ALLOC; ins; apply evEQ.
      by rewrite <- evEQ, SOEQ; ins.
    des; exists to; splits.
      by red in SEQEN; des; red; split; ins; rewrite <- evEQ.
      by rewrite <- hbEQ; ins.
      by ins; rewrite <- rfEQ, <- evEQ in * ; eauto.
  Qed.
  Next Obligation.
    unfold consistent in *; des; splits; ins.
    specialize (SEQEN l); des; exists to; splits; ins. 
    rewrite INC2; ins.
  Qed.
  Next Obligation.
    eexists; splits; ins; red; red in C; desc; splits; ins; rewrite ?SOEQ; 
      try solve [rewrite ?inclusion_restr; eauto].
    - by red in EVa; red in EVb; des; eauto.
    - rewrite <- (DEQN l).
      unfolder; ins; desf; splits; eauto; intro; desf; destruct H1; eauto.
      eexists; splits; eauto.
      unfolder in RFOK; specialize_full RFOK; eauto; desc.
      apply SANE in H3; red in H3; desf.
      assert (L: has_loc a x). 
        by clear - RFOK2; red in RFOK2; destruct x; ins; desf; constructor; exists eq_refl.
      assert (L0: has_loc a x0). 
        by clear - RFOK1; red in RFOK1; destruct x0; ins; desf; constructor; exists eq_refl.
      clear RFOK1 RFOK2.
      rewrite REQ in *; ins; try (by apply set_nonemptyE; eauto).
      unfold set_disjoint in *; ins; eauto.
      enough (x1 = a); desf; eauto using has_loc_unique. 
    - tertium_non_datur (usedLocs l).
      + exists ∅₂.
        assert (EMPT: tup_ev t ∩₁ es ∩₁ has_loc l ⊆₁ ∅).
        { ins; unfolder; ins; des.
          apply REQ in H2; ins.
          by unfolder in H2; eapply H2; eauto.
          by intro CC; unfolder in CC; des; eapply CC; eauto. }
        splits.
        { unfolder; splits; try tauto; ins.
          by apply EMPT in IWa. }
        2: by ins; rewrite EMPT in *; destruct S; ins; edestruct SDOM; ins; eauto.
        unfolder; ins; des.
        unfolder in EMPT; eapply EMPT; splits; eauto.
        red in SANE; des; unfolder in HB_OK; specialize_full HB_OK; eauto; desf.
      + assert (EE: ⦗has_loc l⦘ ⨾ (⦗es⦘ ⨾ tup_rf t ⨾ ⦗es⦘) ⨾ ⦗has_loc l⦘ ≡ ⦗has_loc l⦘ ⨾ tup_rf t ⨾ ⦗has_loc l⦘ ).
        {
          red in SANE; desc.
          rewrite <- !restr_relE; unfolder; split; ins; desf; splits; ins.
          all: specialize (RF_OK _ _ H0); red in RF_OK; desc.
          all: apply REQ; ins; [apply set_nonemptyE; eauto|].
          all: red; ins; enough (x0 = l) by desf; eapply has_loc_unique; eauto. 
        }
        specialize (SEQEN l); des.
        exists to. red in SEQEN; des.
        splits; ins.
        * red; split; ins. red. ins. 
          red in SEQEN2. apply SEQEN2; unfolder in IWa; des; unfolder in IWb; des; unfolder; ins.
        * apply inclusion_trans with (y:=restr_rel (has_loc l) (tup_hb t)); eauto.
          unfolder; ins; des; ins.
        * apply fifo_loc_restr; ins.
            by unfolder; firstorder.
          specialize_full SEQEN1; eauto.
             by rewrite SDOM; unfolder; ins; desf.
          apply fifo_loc_restr in SEQEN1; ins.
          rewrite restr_relE, EE; ins.
   Qed.

End QueueLib.
