Require Import Hahn Omega LibBase Operations.
Require Import Wf IndefiniteDescription.
Require Import Setoid.

Set Implicit Arguments.

Module StackLib.
  Definition is_newstack_loc l e := 
    exists (OP: e.(ev_op) = op_newstack), 
      (eq_rect _ _ e.(ev_res) _ OP) = l.

  Definition is_newstack e := e.(ev_op) = op_newstack.
  Definition is_push e := e.(ev_op) = op_push.
  Definition is_pop e := e.(ev_op) = op_pop.

  Definition is_push_loc l e := 
    exists (OP: e.(ev_op) = op_push), 
      fst (eq_rect _ _ e.(ev_args) _ (f_equal _ OP)) = l.

  Definition is_push_val l v e := 
    exists (OP: e.(ev_op) = op_push), 
      fst (eq_rect _ _ e.(ev_args) _ (f_equal _ OP)) = l /\
      fst (snd (eq_rect _ _ e.(ev_args) _ (f_equal _ OP))) = v.

  Definition is_pop_loc l e := 
    exists (OP: e.(ev_op) = op_pop), 
      fst (eq_rect _ _ e.(ev_args) _ (f_equal _ OP)) = l.
      
  Definition is_pop_val l v e :=
    exists (OP: e.(ev_op) = op_pop), 
      fst (eq_rect _ _ e.(ev_args) _ (f_equal _ OP)) = l /\
      match (eq_rect _ (fun x => x) e.(ev_res) _ (f_equal _ OP)) with
        | Some v' => v = v'
        | _ => False 
      end.

  Definition is_popN l e :=
    exists (OP: e.(ev_op) = op_pop),
      fst (eq_rect _ _ e.(ev_args) _ (f_equal _ OP)) = l /\
      match (eq_rect _ (fun x => x) e.(ev_res) _ (f_equal _ OP)) with
        | None => True
        | _ => False
      end.

  Definition has_loc x := is_push_loc x ∪₁ is_pop_loc x.

  Lemma is_push_impl l v e : is_push_val l v e -> is_push_loc l e.
  Proof. intro H; red in H; red; des; eexists; eauto. Qed.
  
  Lemma is_pop_impl l v e : is_pop_val l v e -> is_pop_loc l e.
  Proof. intro H; red in H; red; des; eexists; eauto. Qed.


  Lemma is_popN_impl l e : is_popN l e -> is_pop_loc l e.
  Proof. intro H; red in H; red; des; eexists; eauto. Qed.

  Lemma pop_loc_unique x l1 l2: is_pop_loc l1 x /\ is_pop_loc l2 x -> l1 = l2.
  Proof.
    ins; des.
    unfold has_loc in *; des.
    red in H; red in H0; des.
      by destruct x; desf; ins; desf; ins; rewrite <- eq_rect_eq in *. 
  Qed.

  Lemma push_loc_unique x l1 l2: is_push_loc l1 x /\ is_push_loc l2 x -> l1 = l2.
  Proof.
    ins; des.
    red in H; red in H0; des.
      by destruct x; desf; ins; desf; ins; rewrite <- eq_rect_eq in *. 
  Qed.

  Lemma has_loc_unique: forall l l' e, has_loc l e -> has_loc l' e -> l = l'.
  Proof.
    unfold has_loc; unfolder; ins; desf; eauto using push_loc_unique, pop_loc_unique.
    all: destruct e; red in H, H0 ; desf; ins; desf.
  Qed.

  Fixpoint is_lifo com l s S := 
    match S with
    | nil => True
    | ev :: S' => is_push_loc l ev /\ is_lifo com l (ev :: s) S' \/
                  match s with 
                  | nil => is_popN l ev /\ is_lifo com l s S'
                  | ev' :: s' => com ev' ev /\ is_lifo com l s' S'
                  end
    end.

  Definition wf (t: tuple) :=
      ⟪EOK: tup_ev t ⊆₁ is_newstack ∪₁ is_push ∪₁ is_pop⟫ /\
      ⟪WF_LOC: forall x, tup_ev t ∩₁ has_loc x
                ⊆₁ codom_rel (⦗is_newstack_loc x⦘ ⨾ (tup_po t ∪ tup_rf t)⁺)⟫ /\
      ⟪ALLOC : forall x a (EVa: tup_ev t a) (Aa: is_newstack_loc x a)
                        b (EVb: tup_ev t b) (Ab: is_newstack_loc x b), a = b⟫.

(* TODO: move this to some util file *)
  Definition finite_prefix A (to : relation A) ls := 
      ⟪RESP: StronglySorted to ls⟫ /\
      ⟪FIN: ls = nil \/ exists el, In el ls /\ forall el', to el' el <-> In el' ls⟫.

  Definition consistent (t : tuple) :=
    ⟪RFOK: tup_rf t ⊆ ⋃l, ⋃v, is_push_val l v × is_pop_val l v⟫ /\
    ⟪ALLOC: forall x a (EVa: tup_ev t a) (Aa: is_newstack_loc x a)
                           b (EVb: tup_ev t b) (Ab: is_newstack_loc x b), a = b⟫ /\
    ⟪RFFUN: functional (tup_rf t)⟫ /\
    ⟪RF1FUN: functional (tup_rf t)⁻¹⟫ /\
    ⟪SOEQ: tup_so t ≡ tup_rf t⟫ /\
    ⟪DEQN: forall l, tup_ev t ∩₁ (is_pop_loc l \₁ codom_rel (tup_rf t)) ⊆₁ is_popN l⟫ /\
    ⟪ENQV: forall l, ⦗is_push_loc l \₁ dom_rel (tup_rf t)⦘ ⨾ tup_hb t ⨾ ⦗is_popN l⦘ ≡ ∅₂⟫ /\
    (* The newqueue bit is out by definition of has_loc l *)
    ⟪SEQEN: forall l, exists to, strict_total_order ( (tup_ev t ∩₁ has_loc l) (*\₁ is_newqueue_loc l*)) to /\
                               restr_rel (has_loc l) (tup_hb t) ⊆ to /\
            forall S, finite_prefix to S -> is_lifo (tup_rf t) l nil S⟫.

  Add Parametric Morphism: is_lifo with signature
    same_relation ==> eq ==> eq ==> eq ==> iff as is_fifo_mor.
  Proof.
    intros com com' EQ l s S.
    revert s.
    induction S; ins.
    split; ins; desf; ins; desf; ins;
    solve [ by left; splits; eauto; apply IHS 
          | by right; splits; eauto; apply IHS
          | by right; splits; eauto; first [apply EQ | apply IHS]].
  Qed.


  Lemma lifo_loc_restr s (com : relation event) l S (COMSUB: com ⊆ ⋃l, ⋃v, is_push_val l v × is_pop_val l v)
        (sOK: forall e, In e s -> has_loc l e):
          is_lifo com l s S <-> is_lifo (<| has_loc l |> ;; com ;; <| has_loc l |>) l s S.
  Proof.
    split.
    - revert sOK; revert s; induction S; ins; des.
      + left; split; ins. 
        apply IHS; ins.
        des; subst.
          by red; left; eauto.
          by apply sOK. 
      + right; case_eq s.
          by ins; subst; des; split; ins; apply IHS.
        ins; rewrite H0 in *;
        des; split; [| apply IHS; ins]; remember H as C; clear HeqC.
        assert (QQQ: has_loc l e)
           by (apply sOK; left; ins).
        assert (WWW: has_loc l a).
        {  apply COMSUB in C; right.
           unfolder in C; des.
           apply is_push_impl in C1.
           apply is_pop_impl in C2.
           enough (l = a0); subst; ins.
           apply push_loc_unique with (x:=e); split; ins.
           specialize (sOK _ (or_introl eq_refl)).
           do 2 (red in sOK); des; ins.
           destruct e; red in sOK, C1; des; ins; desf. }
         by apply COMSUB in H; exists e; split; try exists a; split.
         by apply sOK; right.
    - revert sOK; revert s; induction S; ins; des.
      + left; split; ins; apply IHS; ins.
          des; subst.
          by red; left.
          by apply sOK.
      + right; case_eq s.
          by ins; subst; des; split; ins; apply IHS; ins.
         ins; rewrite H0 in *.
         des; split; [| apply IHS; ins]; remember H as C; clear HeqC.
           by unfolder in C; desf.
           by apply sOK; right.
    Qed.

  Lemma functional_subs A (r r' : relation A) (SUBS: r ⊆ r') (FUNC: functional r') :
    functional r.
  Proof. unfold functional in *; eauto. Qed.

  Lemma restr_subs A (r : relation A) (s : A -> Prop): 
              ⦗s⦘ ⨾ r ⨾ ⦗s⦘ ⊆ r.
  Proof. unfolder; splits; des; ins. Qed.

  Ltac falsify_ := 
    match goal with 
    | H1: is_push_loc _ ?x, H2: is_pop_loc _ ?x |- _ => destruct x; red in H1, H2 ; desf; ins; desf
    end.

  Ltac eqloc_ :=
    match goal with
    | H1: is_push_loc ?a ?x, H2: is_push_loc ?b ?x |- ?a = ?b
            => eapply push_loc_unique; ins; eauto 
    | H1: is_pop_loc ?a ?x, H2: is_pop_loc ?b ?x |- ?a = ?b
            => eapply pop_loc_unique; ins; eauto
    end; subst.


  Program Definition library := 
    {| libspec_ops := op_newstack :: op_push :: op_pop :: nil ;
       libspec_logaddr := stack_address ;
       libspec_allocs e x := is_newstack_loc x e ;
       libspec_requires e x := has_loc x e ;
       libspec_wf := wf ;
       libspec_consistent := consistent |}.
  Next Obligation.
    unfold wf in *; des;
    assert(AA: (tup_ev t ∩₁ has_loc loc) e) by firstorder;
    apply WF_LOC in AA;
    red in AA; des; exists x;
    firstorder; subst; ins.
  Qed.
  Next Obligation.
    unfold wf in *; splits; des; ins.
      by intros x H; red in H; des; apply EOK.
    - rewrite union_restr.
      rewrite (set_interC _ es), set_interA, WF_LOC.
      revert PREFIX; clear.
      generalize (tup_po t ∪ tup_rf t) as r; ins; unfolder in *; ins; desf.
      eexists; split; eauto.
      clear H0; induction H1; eauto 10 using clos_trans.
   - red in EVa, EVb; des; eapply ALLOC; eauto.
  Qed.
  Next Obligation.
    unfold wf in *; des; red; eauto.
  Qed.
  Next Obligation.
    unfold consistent in *; des; red; eauto.
  Qed.
  Next Obligation.
    unfold consistent in *; des; rewrite SOEQ; unfolder; eauto 10 using t_step.
  Qed.
  Next Obligation.
    unfold consistent in *; des; splits; ins.
    all: try repeat match goal with
    | [ H: same_relation _ _ |- _ ] => ins; rewrite <- H; eauto
    | [ H: forall l:stack_address, _ |- _ ] => specialize (H l)
    end.
      by apply ALLOC; ins; apply evEQ.
      by rewrite <- evEQ, SOEQ; ins.
    des; exists to; splits.
      by red in SEQEN; des; red; split; ins; rewrite <- evEQ.
      by rewrite <- hbEQ; ins.
      by ins; rewrite <- rfEQ; eauto.
  Qed.
  Next Obligation.
    unfold consistent in *; des; splits;
    destruct t; ins.
    - split; ins.
      specialize (ENQV l); red in ENQV; des.
      eapply inclusion_trans with (y:=⦗is_push_loc l \₁ dom_rel tup_rf⦘ ⨾ tup_hb ⨾ ⦗is_popN l⦘); ins.
      unfolder; splits; des; ins.
      apply INC2 in H0; ins.
    - specialize (SEQEN l); des; exists to; splits; ins. 
      eapply inclusion_trans; eauto; 
        unfolder; splits; des; ins; eauto.
   Qed.
  Next Obligation.
    eexists; splits; ins; red; red in C; desc; splits; ins; rewrite ?SOEQ; 
      try solve [rewrite ?inclusion_restr; eauto].
    - by red in EVa; red in EVb; des; eauto.
    - rewrite <- (DEQN l).
      unfolder; ins; desf; splits; eauto; intro; desf; destruct H1; eauto.
      eexists; splits; eauto.
      unfolder in RFOK; specialize_full RFOK; eauto; desc.
      apply SANE in H3; red in H3; desf.
      assert (L: has_loc a x). 
        by clear - RFOK2; red in RFOK2; destruct x; ins; desf; constructor; exists eq_refl.
      assert (L0: has_loc a x0). 
        by clear - RFOK1; red in RFOK1; destruct x0; ins; desf; constructor; exists eq_refl.
      clear RFOK1 RFOK2.
      rewrite REQ in *; ins; try (by apply set_nonemptyE; eauto).
      unfold set_disjoint in *; ins; eauto.
      enough (x1 = a); desf; eauto using has_loc_unique. 
    - split; ins.
      unfolder; intros x y HH; desf. 
      specialize (ENQV l); unfolder in ENQV; des.
      apply (ENQV x y); splits; ins.
      intro CONTRA; destruct HH4; desf; eexists; split; eauto.
      eexists; splits; eauto. 
      red in SANE; desc; unfolder in RF_OK; specialize_full RF_OK; eauto.
      apply RFOK in CONTRA; unfolder in CONTRA; des.
      apply is_push_impl in CONTRA1.
      apply is_pop_impl in CONTRA2.
      unfold has_loc, set_union in REQ.
      enough (a = l); [subst | eqloc_].
      rewrite REQ in HH2 |- *; ins; try solve [eapply set_nonemptyE; eauto]. 
      red; ins; desf; try falsify_; enough (x0 = l); [subst | eqloc_]; eauto. 
    - tertium_non_datur (usedLocs l).
      + exists ∅₂.
        assert (EMPT: tup_ev t ∩₁ es ∩₁ has_loc l ⊆₁ ∅).
        { ins; unfolder; ins; des.
          apply REQ in H2; ins.
          by unfolder in H2; eapply H2; eauto.
          by intro CC; unfolder in CC; des; eapply CC; eauto. }
        splits.
        * unfolder; splits; try tauto; ins.
          by apply EMPT in IWa.
        * by red in SANE; desc; rewrite HB_OK; unfolder; unfolder in EMPT; ins; desf; eauto. 
        * by ins; red in H0; des; subst; ins; apply FIN0 in FIN.
      + assert (EE: ⦗has_loc l⦘ ⨾ (⦗es⦘ ⨾ tup_rf t ⨾ ⦗es⦘) ⨾ ⦗has_loc l⦘ ≡ ⦗has_loc l⦘ ⨾ tup_rf t ⨾ ⦗has_loc l⦘ ).
        {
          red in SANE; desc.
          rewrite <- !restr_relE; unfolder; split; ins; desf; splits; ins.
          all: specialize (RF_OK _ _ H0); red in RF_OK; desc.
          all: apply REQ; ins; [apply set_nonemptyE; eauto|].
          all: red; ins; enough (x0 = l) by desf; eapply has_loc_unique; eauto. 
        }
        specialize (SEQEN l); des.
        exists to. red in SEQEN; des.
        splits; ins.
        * red; split; ins. red. ins. 
          red in SEQEN2. apply SEQEN2; unfolder in IWa; des; unfolder in IWb; des; unfolder; ins.
        * apply inclusion_trans with (y:=restr_rel (has_loc l) (tup_hb t)); eauto.
          unfolder; ins; des; ins.
        * apply lifo_loc_restr; ins.
            by unfolder; firstorder.
          specialize_full SEQEN1; eauto.
          apply lifo_loc_restr in SEQEN1; ins.
          rewrite restr_relE, EE; ins.
   Grab Existential Variables. all: auto.
   Qed.

End StackLib.
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    