Require Import String Hahn ZArith Omega.

Set Implicit Arguments.

(** ** Operations *)

(** An address is a pair of a block identifier (a natural number) and
    an offset (an integer). *)

Definition address := (nat * nat)%type.

Definition offset (x : address) (y : nat) : address := 
  let (a,b) := x in (a, b + y).

(** Some C11 operations take as arguments explicit access modes
(a.k.a. memory orders), which are defined below. *)
 
Inductive read_mode := RMna | RMrlx | RMacq | RMsc.
Inductive write_mode := WMna | WMrlx | WMrel | WMsc.
Inductive rmw_mode := UMrlx | UMacq | UMrel | UMacqrel | UMsc.

(** Operations contain a name, a list of types for the arguments, and
a type for the result. *)

Definition vector (Ts: list Type) : Type 
  := fold_right (fun a r => a * r)%type unit Ts.

Record operation : Type := 
  { op_name : string ; 
    op_args: list Type ; 
    op_result : Type ; 
  }.

Notation "'CONSTRUCTOR' op : z 'end'" := 
  ({| op_name := op ;
     op_args := nil ; 
     op_result := z |}) 
    (at level 10, op at level 0, z at level 0, only parsing).

Notation "'OPER' op : x * .. * y -> z 'end'" := 
  ({| op_name := op ;
     op_args := @cons Type x (.. (@cons Type y nil) .. ) ; 
     op_result := z |}) 
    (at level 10, op at level 0, x at level 0, y at level 0, z at level 0,
     only parsing).

Record event := 
  { ev_id : nat
  ; ev_op : operation
  ; ev_args : vector (op_args ev_op)
  ; ev_res : op_result ev_op }.

(** Now we define some example of operations.
We start with a generic allocator, returning a fresh block. *)

Definition alloc_block := 
  CONSTRUCTOR "__alloc_block" : nat 
  end.

(** Next, we define the types of C11 operations: loads, stores, 
various kinds of read-modify-write instructions, and fences. *)

Definition C11_load := 
  OPER "load_atomic_explicit" : address * read_mode -> Z 
  end.

Definition C11_store := 
  OPER "store_atomic_explicit" : address * Z * write_mode -> unit 
  end.

Definition C11_cas := 
  OPER "atomic_compare_exchange_strong_explicit" : 
    address * Z * Z * rmw_mode * read_mode -> Z
  end.

Definition C11_fadd := 
  OPER "atomic_fetch_add_explicit" : 
    address * Z * rmw_mode -> Z 
  end.

Definition C11_xchg := 
  OPER "atomic_exchange_explicit" : 
    address * Z * rmw_mode -> Z
  end.

Definition C11_fence :=
  OPER "atomic_signal_fence" : rmw_mode -> unit
  end.

(** A simple register library supports register allocation, reading
and writing operations. *)

Definition reg_alloc :=
  CONSTRUCTOR "reg_alloc" : address
  end.

Definition reg_read :=
  OPER "reg_read" : address -> Z
  end.

Definition reg_write :=
  OPER "reg_write" : address * Z -> unit
  end.

(** The basic mutex lock library supports four operations: (1)
allocating a new lock, (2) initialising a lock at a given already
allocated address, (3) acquiring and (4) releasing the lock. *)

Definition op_newlock := 
  CONSTRUCTOR "newlock" : address
  end.

Definition op_initlock := 
  OPER "initlock" : address -> unit
  end.

Definition op_lock := 
  OPER "lock" : address -> unit
  end.

Definition op_unlock := 
  OPER "unlock" : address -> unit
  end.

(** The queue library supports three operations: 
allocating a queue, enqueuing an element to the end of a queue, 
and dequeuing an element from the beginning of the queue. *)

Definition queue_address := (nat * address)%type.

Definition op_newqueue := 
  CONSTRUCTOR "newqueue" : queue_address
  end.

Definition op_enqueue := 
  OPER "enqueue" : queue_address * Z -> unit
  end.

Definition op_dequeue := 
  OPER "dequeue" : queue_address -> option Z
  end.

(** The stack library supports three operations: 
allocating a stack, pushing a value to the top of the stack,
and popping a value from the top of the stack.
*)
Definition stack_address := (nat * address)%type.
Definition op_newstack := 
  CONSTRUCTOR "newstack" : stack_address
  end.

Definition op_push := 
  OPER "enqueue" : stack_address * Z -> unit
  end.

Definition op_pop := 
  OPER "dequeue" : stack_address -> option Z
  end.

(** The exchanger library supports two operations: 
allocating an exchanger, and requesting to exchange a value.
*)

Definition op_newexchanger := 
  CONSTRUCTOR "newexchanger" : nat
  end.

Definition op_exchange := 
  OPER "exchange" : nat * Z -> option Z
  end.

