Require Import Hahn Operations Language LangSemantics ExpandFcall.
Require Import LibBase LibC11 LibLock.
Require Import ZArith Coq.micromega.Lia IndefiniteDescription.

Set Implicit Arguments.

Ltac contra :=
  solve [exfalso; repeat match goal with 
                        | e : event |- _ => destruct e; simpl in * 
                        | e : unit |- _ => destruct e
                        | e : (_ * _)%type |- _ => destruct e; simpl in *
                        | e : ?x = ?x |- _ => rewrite (UIP_refl _ _ e) in *; clear e; simpl in *
                        | H : ?P |- _ => match type of P with Prop => red in H; desf; simpl in * end
                   end; desf; auto].

Lemma event_of_alloc_block res a :
  event_of_call alloc_block tt res a ->
  ⟪LIBC: lib_events RC11.library a⟫ /\ 
  ⟪ALLOC: RC11.is_alloc_ res a⟫.
Proof.
  unfold event_of_call; destruct a; ins; desf; ins; splits; red; ins.
  all: repeat constructor; exists eq_refl; ins.
Qed.

Lemma event_of_load loc mode loadv a :
  event_of_call C11_load (loc, (mode, tt)) loadv a ->
  ⟪LIBC: lib_events RC11.library a⟫ /\ 
  ⟪LD: RC11.is_load a⟫ /\ ⟪LDLOC: RC11.has_loc loc a⟫ /\
  ⟪LDVAL: RC11.has_rval loadv a⟫ /\ ⟪LDMODE: RC11.has_read_mode mode a⟫.
Proof.
  unfold event_of_call; destruct a; ins; desf; ins; splits; red; ins.
  all: repeat constructor; exists eq_refl; ins.
Qed.

Lemma event_of_store loc val mode r a :
  event_of_call C11_store (loc, (val, (mode, tt))) r a ->
  r = tt /\
  ⟪LIBC: lib_events RC11.library a⟫ /\ 
  ⟪ST: RC11.is_store a⟫ /\ ⟪STLOC: RC11.has_loc loc a⟫ /\
  ⟪STVAL: RC11.has_wval val a⟫ /\ ⟪STMODE: RC11.has_write_mode mode a⟫.
Proof.
  unfold event_of_call; destruct a; ins; desf; destruct ev_res; ins; splits; try red; ins.
  all: repeat constructor; exists eq_refl; ins.
Qed.

Lemma event_of_cas loc expval newval rmwmode rmode r a : 
  event_of_call C11_cas (loc, (expval, (newval, (rmwmode, (rmode, tt))))) r a ->
  ⟪rEQ: r = expval⟫ /\
  ⟪LIBC: lib_events RC11.library a⟫ /\ 
  ⟪CASOK: RC11.is_rmw a⟫ /\ 
  ⟪CASLOC: RC11.has_loc loc a⟫ /\
  ⟪CASRVAL: RC11.has_rval expval a⟫ /\ ⟪CASRMODE: RC11.has_read_mode (read_mode_of_rmw rmwmode) a⟫ /\
  ⟪CASWVAL: RC11.has_wval newval a⟫ /\ ⟪CASWMODE: RC11.has_write_mode (write_mode_of_rmw rmwmode) a⟫ 
  \/ ⟪CASVALF: r <> expval⟫ /\
     ⟪LIBC: lib_events RC11.library a⟫ /\ 
     ⟪CASFAIL: RC11.is_load a⟫ /\ 
     ⟪CASLOC: RC11.has_loc loc a⟫ /\
     ⟪CASMODE: RC11.has_read_mode rmode a⟫ /\
     ⟪CASRVAL: RC11.has_rval r a⟫.
Proof.
  unfold event_of_call; destruct a; ins; desf; ins.
  destruct (Z.eq_dec ev_res expval); [left|right]; splits; ins.
  all: try (repeat constructor; exists eq_refl; done).
Qed.

Lemma event_of_newlock r a :
  event_of_call op_newlock (tt) r a ->
  ⟪LIBC: lib_events LockLib.library a⟫ /\ 
  ⟪LOCK: LockLib.is_newlock_ok r a⟫.
Proof.
  unfold event_of_call; destruct a; ins; desf; destruct ev_res; ins; splits; try red; ins.
  all: repeat constructor; exists eq_refl; ins.
Qed.

Lemma event_of_lock loc r a :
  event_of_call op_lock (loc, tt) r a ->
  r = tt /\
  ⟪LIBC: lib_events LockLib.library a⟫ /\ 
  ⟪LOCK: LockLib.is_lock a⟫ /\ ⟪LOCKLOC: LockLib.is_ok loc a⟫.
Proof.
  unfold event_of_call; destruct a; ins; desf; destruct ev_res; ins; splits; try red; ins.
  all: repeat constructor; exists eq_refl; ins.
Qed.

Lemma event_of_unlock loc r a :
  event_of_call op_unlock (loc, tt) r a ->
  r = tt /\
  ⟪LIBC: lib_events LockLib.library a⟫ /\ 
  ⟪LOCK: LockLib.is_unlock a⟫ /\ ⟪LOCKLOC: LockLib.is_ok loc a⟫.
Proof.
  unfold event_of_call; destruct a; ins; desf; destruct ev_res; ins; splits; try red; ins.
  all: repeat constructor; exists eq_refl; ins.
Qed.


Ltac event_ofs := 
  repeat 
    match goal with 
    | H : event_of_call _ _ _ _ |- _ => 
      first [ apply event_of_alloc_block in H 
            | apply event_of_load in H
            | apply event_of_store in H
            | apply event_of_cas in H 
            | apply event_of_newlock in H 
            | apply event_of_lock in H 
            | apply event_of_unlock in H ]
    end; desc; clarify.

Lemma set_equiv_inr : forall A x (S S': A -> Prop), S ≡₁ S' -> S x -> S' x.
Proof.
  firstorder.
Qed.

Lemma set_equiv_inl : forall A x (S S': A -> Prop), S ≡₁ S' -> S' x -> S x.
Proof.
  firstorder.
Qed.
     
Lemma show_restr_rel A (cond: A -> Prop) (R : relation A) x y: 
  cond x -> cond y -> R x y -> restr_rel cond R x y.
Proof.
  vauto.
Qed.

Lemma same_relation_inr : forall A x y (R R': relation A), R ≡ R' -> R x y -> R' x y.
Proof.
  firstorder.
Qed.

Lemma same_relation_inl : forall A x y (R R': relation A), R ≡ R' -> R' x y -> R x y.
Proof.
  firstorder.
Qed.

Lemma strict_partial_order_restr A cond (R: relation A): 
  strict_partial_order R -> strict_partial_order (restr_rel cond R).
Proof. firstorder. Qed.



Ltac rewrite_gs :=
  repeat first 
         [match goal with
          | H : _ ≡ _ |- _ => clear H 
          | H : set_disjoint _ _ |- _ => clear H
          | H : _ ≡₁ _ |- _ => rewrite H end
         | rewrite set_le, set_bunion_union_l, set_bunion_eq (* last loop interation *)];
  autorewrite with hahn hahn_full; splits; ins.

Ltac clear_gpos :=
  repeat match goal with
  | H : _ ≡ _ |- _ => clear H
  | H : set_disjoint _ _ |- _ => clear H
  | x : relation event |- _ => clear x
  end.

Ltac rewrite_gpos :=
  repeat first 
         [match goal with
          | H : ?x ≡ _ |- _ => rewrite H in *; try (clear x H || autorewrite with hahn in H)
          | H : _ ≡₁ _ |- _ => rewrite set_le, set_bunion_union_l, set_bunion_eq in H
          | H : ?x ≡₁ _ |- _ => rewrite H in *; try (clear x H || autorewrite with hahn in H) end
         | rewrite set_le, set_bunion_union_l, set_bunion_eq (* last loop interation *)];
  autorewrite with hahn in *; ins.
  
Notation "'IMPL' x , .. , y " := 
  (env_compose (env_of_impl x) (.. (env_compose (env_of_impl y) (fun _ => None)) .. )) 
    (at level 10, x at level 0, y at level 0).

Fixpoint expr_ops A As (e: expr A As) : event -> Prop := 
  match e with 
  | Evalue v => ∅
  | Ecall op args => fun a =>
     (exists (OP: ev_op a = op), 
         ev_args a = eq_rect args (f_equal _ (eq_sym OP)))
  | Elet e1 e2 => expr_ops e1 ∪₁ ⋃₁ v, expr_ops (e2 v)   
  | Eloop e => expr_ops e
  | Ebreak n v EQ => ∅
  | Epar e1 e2 => expr_ops e1 ∪₁ expr_ops e2
  end.

Lemma op_approx A As (e: expr A As) r g po (SEM: expr_sem e r g po) :
  g ⊆₁ expr_ops e.
Proof.
  revert r g po SEM; induction e; ins; unfold event_of_call, idom_seq, idom_loop, idom_par in *; 
    desf; ins; desf; clear_gpos; rewrite_gpos.
  all: unfolder in *; ins; desf; eauto.
  unfolder; ins; desf; exists eq_refl; ins.
Qed.


(** Expression semantics *)

Fixpoint expr_sem2 A As (e: expr A As) : 
  return_type A As -> (event -> Prop) -> relation event -> Prop :=
  match e with 
  | Evalue v => fun r g gpo => r = RTbase v /\ g ≡₁ ∅
  | Ecall op args => fun r g gpo =>
     (exists a (OP: ev_op a = op), 
          << ARGS: ev_args a = eq_rect args (f_equal _ (eq_sym OP)) >>
       /\ << gEQ: g ≡₁ eq a >>
       /\ << rEQ: r = RTbase (eq_rect (ev_res a) OP) >>)
  | Elet e1 e2 => fun r g gpo =>
       (exists v g1 g2, 
           expr_sem2 e1 (RTbase v) g1 gpo 
           /\ expr_sem2 (e2 v) r g2 gpo 
           /\ set_disjoint (set_collect ev_id g1) (set_collect ev_id g2) 
           /\ g ≡₁ g1 ∪₁ g2 /\ g1 × g2 ⊆ gpo)
    \/ (exists C Cs (EQ : _ = C :: Cs) (rt : return_type C Cs), 
           expr_sem2 e1 (RTcons EQ rt) g gpo /\ r = RTcons EQ rt)
  | Eloop e => 
    fun r g gpo => 
      exists (n: nat) B Bs EQ rt gn,
        << NLAST: forall i (LT: i < n), expr_sem2 e (RTbase tt) (gn i) gpo >> /\
        << LAST: expr_sem2 e (@RTcons _ _ B Bs EQ rt) (gn n) gpo >> /\
        << rEQ : r = 
                 let mEQ := eq_sym (proj1 (cons_inj EQ)) in
                 let nEQ := eq_sym (proj2 (cons_inj EQ)) in
                 (eq_rect (P:=fun As => return_type _ As) 
                    (eq_rect (P:=fun A => return_type A Bs) rt mEQ) nEQ) >> /\
        << gCOMPAT : forall i j (LT: i < j) (LEQ: j <= n), 
                    set_disjoint (set_collect ev_id (gn i)) (set_collect ev_id (gn j)) >> /\
        << gEQ : g ≡₁ ⋃₁ i <= n, gn i >> /\
        << gEQ : (⋃i <= n, ⋃j < i, gn j × gn i) ⊆ gpo >>
  | Ebreak n v EQ => fun r g gpo => r = RTconsn n v EQ /\ g ≡₁ ∅
  | Epar e1 e2 => fun r g gpo => 
      exists v1 g1 v2 g2, 
        expr_sem2 e1 (RTbase v1) g1 gpo /\ expr_sem2 e2 (RTbase v2) g2 gpo
        /\ set_disjoint (set_collect ev_id g1) (set_collect ev_id g2) 
        /\ g ≡₁ g1 ∪₁ g2 /\ r = RTbase (v1, v2)
  end.

Lemma op_approx2 A As (e: expr A As) r g po (SEM: expr_sem e (Some r) g po) 
      po' (INCL: po ⊆ po'):
  expr_sem2 e r g po'.
Proof.
  revert r g po SEM INCL; induction e; ins; 
    unfold event_of_call, idom_seq, idom_loop, idom_loop_helper, idom_par, idom_par_helper in *; 
    desf; ins; desf.
  * by eexists; exists eq_refl; splits; ins.
  * specialize_full IHe; eauto; [by firstorder|].
    specialize_full H; eauto; [by firstorder|].
    by left; do 3 eexists; splits; eauto; firstorder.
  * specialize_full IHe; eauto. 
    by right; repeat eexists; splits; eauto; firstorder.
  * assert (M:=INCL); rewrite poEQ in M. 
    eapply IHe in LAST; eauto; [|by clear - M; firstorder].
    do 6 eexists; splits; eauto. 
    ins; eapply IHe; eauto; clear -M LT; firstorder.
    rewrite <- M; clear; firstorder.
  * specialize_full IHe1; eauto; [by firstorder|].
    specialize_full IHe2; eauto; [by firstorder|].
    do 4 eexists; splits; eauto; firstorder.
Qed.
    
Lemma fsupp_immediate_pred A (R: relation A) 
    (FSUPP: fsupp R) (IRR: irreflexive R) (TRANS: transitive R) 
    x (NMIN: ~ min_elt R x): 
  exists x', immediate R x' x.
Proof. 
  clarify_not.
  apply fsupp_imm_t, ct_end in NMIN; ins; destruct NMIN; desf; eauto.
Qed.

Definition min_elt_restr A cond (R: relation A) a := cond a /\ forall b, cond b -> ~ R b a.

Lemma min_elt_restr_in A cond R a: @min_elt_restr A cond R a -> cond a.
Proof. 
  firstorder.
Qed.

Lemma fsupp_immediate_pred_restr A (R: relation A) cond (FSUPP: fsupp (restr_rel cond R)) 
    (IRR: irreflexive (restr_rel cond R)) (TRANS: transitive (restr_rel cond R)) 
    x (Cx: cond x) (NMIN: ~ min_elt_restr cond R x): 
  exists x', immediate (restr_rel cond R) x' x.
Proof. 
  clarify_not.
  eapply show_restr_rel, fsupp_imm_t, ct_end in NMIN0; eauto.
  ins; destruct NMIN0; desf; eauto.
Qed.

  
Lemma partial_order_is_total_sufficient A (R: relation A) cond 
    (IRR: irreflexive (restr_rel cond R))
    (TRANS: transitive (restr_rel cond R))
    (FSUPP: fsupp (restr_rel cond R))
    (FUNC: functional (immediate (restr_rel cond R)))
    (MIN: cond ≡₁ ∅ \/ exists m, min_elt_restr cond R ≡₁ eq m):
  is_total cond R.
Proof.
  desf; [firstorder | ].
  assert (Rm: forall x, x <> m -> cond x -> R m x).
  { induction x using (well_founded_induction (fsupp_well_founded FSUPP IRR TRANS)).
    intros NEQ Cx.
    assert (X: ~ min_elt_restr cond R x); clarify_not.
      by intro; apply NEQ; symmetry; apply MIN.
    tertium_non_datur (n = m); desf.
    eapply inclusion_restr, TRANS; vauto.
    red; splits; eauto.
    by eapply H; vauto.
    forward apply (set_equiv_inl m MIN); firstorder.
  }
  red; ins.
  tertium_non_datur (R b a) as [?|notRba]; eauto.
  left.
  generalize dependent b.
  induction a using (well_founded_induction (fsupp_well_founded FSUPP IRR TRANS)).
  rename H into IHa; ins.
  tertium_non_datur (a = m) as [?|AM]; desf; eauto.
  forward eapply fsupp_immediate_pred_restr with (x := a) as notMINa; eauto; desf.
    by forward apply (set_equiv_inl m MIN); firstorder.  
  forward eapply (IHa x'); eauto.
    1-3: by unfolder in notMINa; try intro; desf.
    eby intro; unfolder in notMINa; desf; eapply notRba, TRANS.
  intro Rx'b.
  eapply (show_restr_rel cond) in Rx'b; eauto.
    2: by unfolder in notMINa; desf.
  hahn_rewrite (fsupp_imm_t FSUPP IRR TRANS) in Rx'b.
  rewrite clos_trans_t1n_iff in Rx'b.
  inv Rx'b.
    eby exfalso; eapply NEQ, FUNC.
  rewrite (FUNC _ _ _ H notMINa) in *.
  rewrite <- clos_trans_t1n_iff in H0.
  hahn_rewrite <- (fsupp_imm_t FSUPP IRR TRANS) in H0.
  unfolder in H0; desf.
Qed.


Lemma immediate_clos_trans A (R: relation A) (IRR: acyclic R) (FUN: functional R): immediate R⁺ ≡ R.
Proof.
  unfolder; split; ins; desf.
  * inv H; eauto. 
    exfalso; eauto.
  * split; ins; vauto.
    rewrite clos_trans_t1n_iff in R1; inv R1.
    all: rewrite (FUN _ _ _ H H0), <- clos_trans_t1n_iff in *; eauto using t_trans.
Qed.

Lemma clos_trans_restr A (R: relation A) cond (CODOM: codom_rel (<|cond|> ;; R) ⊆₁ cond): 
  restr_rel cond R⁺ ≡ (restr_rel cond R)⁺.
Proof.
  split; red; ins.
    2: by apply restr_ct.
  unfolder in H; desf.
  rewrite clos_trans_t1n_iff in H; induction H; vauto.
  forward apply (CODOM y); [by unfolder; eauto | ins].
  eapply t_trans; eauto; try apply IHclos_trans_1n.
  unfolder; eauto using t_step.
Qed.

Corollary conditional_functional_choice:
forall (A B : Type) (R: A -> B -> Prop) (cond : A -> Prop),
    ~ cond ≡₁ ∅ ->
    (forall (x : A) (COND: cond x), exists y : B, R x y) ->
    (exists f : A -> B, forall (x : A) (COND: cond x), R x (f x)).
Proof.
 ins.
 forward eapply functional_choice 
         with (A := {x : A | cond x}) (B := B) (R := fun x y => R (proj1_sig x) y) as F; desf.
   by destruct x; ins; eauto.
 unfolder in H; clarify_not.
 exists (fun x => match excluded_middle_informative (cond x) with
                    | left C => f (exist _ x C)
                    | _      => f (exist _ n H)
                  end).
 ins; desf.
 specialize (F (exist _ x c)); ins.
Qed. 
    
Lemma finite_codomain_implies_finite_domain A B (P: A -> Prop) (f: A -> B)
        (INJ: forall x y (Px: P x) (Py: P y), f x = f y -> x = y)
        codom (FIN: forall x, P x -> In (f x) codom): 
      exists dom, forall x, P x -> In x dom.
Proof.
  tertium_non_datur (P ≡₁ ∅) as [EMP | NEMP].
    by exists nil; intros x H; apply EMP in H.
  remember (fun b => match excluded_middle_informative (exists a, P a /\ f a = b) with
                       | left EX => proj1_sig (constructive_indefinite_description _ EX)
                       | _       => n
                     end) as finv.
  exists (map finv codom); ins.
  apply in_map_iff.
  exists (f x); split; eauto; desf; desf.
  by match goal with H : _ |- proj1_sig ?t = _ => destruct t end; ins; desf; eauto.
  by firstorder.
Qed.
    
Lemma well_founded_has_min A (R: relation A) (WF: well_founded R) x:
  exists m, R^* m x /\ min_elt R m.
Proof.
  induction x using (well_founded_induction WF).
  tertium_non_datur (min_elt R x).
    by eexists; vauto.
  specialize (H _ H0); desf.
  eexists; split; vauto.
Qed.
  
    
Corollary well_founded_nonempty_has_min A (R: relation A) cond (WF: well_founded (restr_rel cond R)):
  cond ≡₁ ∅ \/ exists m, min_elt_restr cond R m.
Proof.
  tertium_non_datur (cond ≡₁ ∅); eauto.
  eapply well_founded_has_min with (x := n) in WF; desf.
  assert (cond m).
  { apply clos_rt_rt1n_iff in WF; inv WF.
    unfolder in H1; desf.
  }
  right; exists m; split; try done.
  ins; intro; eapply WF0.
  red; splits; eauto.
Qed.
 
